"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:
"""

import os

import cv2
import numpy as np
import pyvista as pv
import SimpleITK as sitk

import vtk
from vtk.util import numpy_support


class ContourToDiscreteMesh(object):
    def __init__(self, contour_position=None, contour_pixel=None, spacing=None, origin=None, dimensions=None,
                 matrix=None, plane='Axial', mask=None):
        self.contour_position = contour_position
        self.contour_pixel = contour_pixel
        self.spacing = spacing
        self.origin = origin
        self.dimensions = dimensions
        self.plane = plane

        self.mask = mask

        if matrix is None:
            self.matrix = np.identity(3)
        else:
            self.matrix = matrix

        if self.contour_pixel is None and self.mask is None:
            self.convert_to_pixel_spacing()

        if self.mask is None:
            self.compute_mask()

    def convert_to_pixel_spacing(self):
        matrix = np.identity(3, dtype=np.float32)
        matrix[0, :] = self.matrix[0, :] / self.spacing[0]
        matrix[1, :] = self.matrix[1, :] / self.spacing[1]
        matrix[2, :] = self.matrix[2, :] / self.spacing[2]

        conversion_matrix = np.identity(4, dtype=np.float32)
        conversion_matrix[:3, :3] = matrix
        conversion_matrix[:3, 3] = np.asarray(self.origin).dot(-matrix.T)

        self.contour_pixel = []
        for ii, pos in enumerate(self.contour_position):
            p_concat = np.concatenate((pos, np.ones((pos.shape[0], 1))), axis=1)
            self.contour_pixel += [p_concat.dot(conversion_matrix.T)[:, :3]]

    def compute_mask(self):
        hold_mask = np.zeros([self.dimensions[0], self.dimensions[1], self.dimensions[2]], dtype=np.uint8)
        if self.plane == 'Axial':
            slice_check = np.zeros(self.dimensions[0])
            for c in self.contour_pixel:
                contour_stacked = np.vstack((c[:, 0:2], c[0, 0:2]))
                new_contour = np.array([contour_stacked], dtype=np.int32)
                image = np.zeros([self.dimensions[1], self.dimensions[2]], dtype=np.uint8)
                cv2.fillPoly(image, new_contour, 1)

                slice_num = int(np.round(c[0, 2]))
                if slice_check[slice_num] == 0:
                    hold_mask[slice_num, :, :] = image
                    slice_check[slice_num] = 1
                else:
                    hold_mask[slice_num, :, :] = hold_mask[slice_num, :, :] + image
            self.mask = (hold_mask > 0).astype(np.uint8)

        elif self.plane == 'Coronal':
            slice_check = np.zeros(self.dimensions[1])
            for c in self.contour_pixel:
                stack_1 = np.vstack((c[:, 0], c[:, 2])).T
                stack_2 = np.asarray([c[0, 0], c[0, 2]])
                contour_stacked = np.vstack((stack_1, stack_2))
                new_contour = np.array([contour_stacked], dtype=np.int32)
                image = np.zeros([self.dimensions[0], self.dimensions[2]], dtype=np.uint8)
                cv2.fillPoly(image, new_contour, 1)

                slice_num = int(np.round(c[0, 1]))
                if slice_check[slice_num] == 0:
                    hold_mask[:, slice_num, :] = image
                    slice_check[slice_num] = 1
                else:
                    hold_mask[:, slice_num, :] = hold_mask[:, slice_num, :] + image
            self.mask = (hold_mask > 0).astype(np.uint8)

        else:
            slice_check = np.zeros(self.dimensions[2])
            for c in self.contour_pixel:
                contour_stacked = np.vstack((c[:, 1:], c[0, 1:]))
                new_contour = np.array([contour_stacked], dtype=np.int32)
                image = np.zeros([self.dimensions[0], self.dimensions[1]], dtype=np.uint8)
                cv2.fillPoly(image, new_contour, 1)

                slice_num = int(np.round(c[0, 0]))
                if slice_check[slice_num] == 0:
                    hold_mask[:, :, slice_num] = image
                    slice_check[slice_num] = 1
                else:
                    hold_mask[:, :, slice_num] = hold_mask[:, :, slice_num] + image
            self.mask = (hold_mask > 0).astype(np.uint8)

    def compute_mesh(self, discrete=False, smoothing_iterations=20, smoothing_relaxation=.5, smoothing_distance=1):
        label = numpy_support.numpy_to_vtk(num_array=self.mask.ravel(),
                                           deep=True, 
                                           array_type=vtk.VTK_FLOAT)

        img_vtk = vtk.vtkImageData()
        img_vtk.SetDimensions(self.dimensions[2], self.dimensions[1], self.dimensions[0])
        img_vtk.SetSpacing(self.spacing)
        img_vtk.SetOrigin(self.origin)
        img_vtk.SetDirectionMatrix(self.matrix.flatten(order='F'))
        img_vtk.GetPointData().SetScalars(label)

        pad_filter = vtk.vtkImageConstantPad()
        pad_filter.SetInputData(img_vtk)
        pad_filter.SetConstant(0)

        old_extent = img_vtk.GetExtent()
        new_extent = [old_extent[0] - 1, old_extent[1] + 1,
                      old_extent[2] - 1, old_extent[3] + 1,
                      old_extent[4] - 1, old_extent[5] + 1]
        pad_filter.SetOutputWholeExtent(new_extent)
        pad_filter.Update()
        pad_image = pad_filter.GetOutput()

        if discrete:
            vtk_mesh = vtk.vtkDiscreteMarchingCubes()
            vtk_mesh.SetInputData(pad_image)
            vtk_mesh.GenerateValues(1, 1, 1)
            vtk_mesh.Update()

            return pv.PolyData(vtk_mesh.GetOutput())

        else:
            # Uses VTK surface nets 3d
            img = pv.ImageData(pad_image)
            mesh = img.contour_labels(smoothing=True,
                                       output_mesh_type='triangles',
                                       smoothing_iterations=smoothing_iterations,
                                       smoothing_relaxation=smoothing_relaxation,
                                       smoothing_distance=smoothing_distance)

            return mesh


class ContourToMask(object):
    def __init__(self, contour_position=None, contour_pixel=None, spacing=None, origin=None, dimensions=None,
                 matrix=None, plane='Axial'):
        self.contour_position = contour_position
        self.contour_pixel = contour_pixel
        self.spacing = spacing
        self.origin = origin
        self.dimensions = dimensions
        self.matrix = matrix
        self.plane = plane

        self.mask = None

    def create_mask(self):
        if self.contour_pixel is None:
            self.convert_to_pixel_spacing()

        self.compute_mask()

    def convert_to_pixel_spacing(self):
        sitk_image = sitk.Image([int(dim) for dim in self.dimensions], sitk.sitkUInt8)
        matrix_flat = self.matrix[0:3, 0:3].flatten(order='F')
        sitk_image.SetDirection([float(mat) for mat in matrix_flat])
        sitk_image.SetOrigin(self.origin)
        sitk_image.SetSpacing(self.spacing)

        self.contour_pixel = [[]] * len(self.contour_position)
        for ii, contours in enumerate(self.contour_position):
            self.contour_pixel[ii] = [sitk_image.TransformPhysicalPointToContinuousIndex(contour) for contour in contours]

    def compute_mask(self):
        hold_mask = np.zeros([self.dimensions[0], self.dimensions[1], self.dimensions[2]], dtype=np.uint8)
        if self.plane == 'Axial':
            slice_check = np.zeros(self.dimensions[0])
            for c in self.contour_pixel:
                contour_stacked = np.vstack((c[:, 1:], c[0, 1:]))
                new_contour = np.array([contour_stacked], dtype=np.int32)
                image = np.zeros([self.dimensions[1], self.dimensions[2]], dtype=np.uint8)
                cv2.fillPoly(image, new_contour, 1)

                slice_num = int(np.round(c[0, 0]))
                if slice_check[slice_num] == 0:
                    hold_mask[slice_num, :, :] = image
                    slice_check[slice_num] = 1
                else:
                    hold_mask[slice_num, :, :] = hold_mask[slice_num, :, :] + image
            self.mask = (hold_mask > 0).astype(np.uint8)

        elif self.plane == 'Coronal':
            slice_check = np.zeros(self.dimensions[1])
            for c in self.contour_pixel:
                stack_1 = np.vstack((c[:, 0], c[:, 2])).T
                stack_2 = np.asarray([c[0, 0], c[0, 2]])
                contour_stacked = np.vstack((stack_1, stack_2))
                new_contour = np.array([contour_stacked], dtype=np.int32)
                image = np.zeros([self.dimensions[0], self.dimensions[2]], dtype=np.uint8)
                cv2.fillPoly(image, new_contour, 1)

                slice_num = int(np.round(c[0, 1]))
                if slice_check[slice_num] == 0:
                    hold_mask[:, slice_num, :] = image
                    slice_check[slice_num] = 1
                else:
                    hold_mask[:, slice_num, :] = hold_mask[:, slice_num, :] + image
            self.mask = (hold_mask > 0).astype(np.uint8)

        else:
            slice_check = np.zeros(self.dimensions[2])
            for c in self.contour_pixel:
                contour_stacked = np.vstack((c[:, 1:], c[0, 1:]))
                new_contour = np.array([contour_stacked], dtype=np.int32)
                image = np.zeros([self.dimensions[0], self.dimensions[1]], dtype=np.uint8)
                cv2.fillPoly(image, new_contour, 1)

                slice_num = int(np.round(c[0, 0]))
                if slice_check[slice_num] == 0:
                    hold_mask[:, :, slice_num] = image
                    slice_check[slice_num] = 1
                else:
                    hold_mask[:, :, slice_num] = hold_mask[:, :, slice_num] + image
            self.mask = (hold_mask > 0).astype(np.uint8)


class ModelToMask:
    """
    Converts a 3D model/s into a mask. The mask can either be an empty array (default) or a binary mask of the model/s.
    The mask is given a spacing buffer of 5 indexes on each side x, y, z.
    """
    def __init__(self, models, origin=None, spacing=None, dims=None, slice_locations=None, matrix=None,
                 empty_array=True, convert=True):
        """

        Parameters
        ----------
        models - List of all models
        empty_array -
        convert
        """
        self.models = models
        self.empty_array = empty_array

        self.spacing = spacing
        self.origin = origin
        self.dims = dims
        self.slice_locations = slice_locations

        if matrix is None:
            self.matrix = np.identity(4)
        else:
            self.matrix = matrix

        self.bounds = None

        self.contours = []
        self.mask = None

        if convert:
            self.compute_bounds()
            self.compute_contours()
            self.compute_mask()

    def compute_bounds(self):
        """
        Computes the boundary for the mask using the model/s bounds. The boundary is the min/max x, y, z combination of
        all the model/s bounds. Default spacing options are [1, 1, 3] or [1, 1, 5] depending on the z axis bound. If
        the bounds are too large then nothing is computed with the assumption that the models are not from the same
        image and the models should have their own independent mask.
        Returns
        -------

        """
        model_bounds = [model.GetBounds() for model in self.models]
        model_min = np.min(model_bounds, axis=0)
        model_max = np.max(model_bounds, axis=0)

        model_min_max = [model_min[0], model_max[1], model_min[2], model_max[3], model_min[4], model_max[5]]

        if model_min_max[1] - model_min_max[0] < 512 and model_min_max[3] - model_min_max[2] < 512:
            if model_min_max[5] - model_min_max[4] < 450:
                self.spacing = [1, 1, 3]

            elif model_min_max[5] - model_min_max[4] < 750:
                self.spacing = [1, 1, 5]

        if self.spacing is not None:
            self.bounds = [int(model_min_max[0] - 5 * self.spacing[0]), int(model_min_max[1] + 5 * self.spacing[0]),
                           int(model_min_max[2] - 5 * self.spacing[1]), int(model_min_max[3] + 5 * self.spacing[1]),
                           int(model_min_max[4] - 5 * self.spacing[2]), int(model_min_max[5] + 5 * self.spacing[2])]
            self.origin = [self.bounds[0], self.bounds[2], self.bounds[4]]

            self.slice_locations = [i for i in range(self.bounds[4], self.bounds[5], self.spacing[2])]
            self.dims = [len(self.slice_locations), self.bounds[3] - self.bounds[2] + 1, self.bounds[1] - self.bounds[0] + 1]

    def compute_contours(self):
        """
        Compute the contours along the z-axis for all models.
        Returns
        -------

        """
        for model in self.models:
            com = model.center
            org_bounds = model.GetBounds()

            model_contours = []
            for s in self.slice_locations:
                if org_bounds[4] < s < org_bounds[5]:
                    hold_contour = model.slice(normal='z', origin=[com[0], com[1], s])
                    model_contours.append((np.asarray(hold_contour.points)[:, 0:2] -
                                     (self.bounds[0], self.bounds[2])) / (self.spacing[0:2]))
                else:
                    model_contours.append([])

            self.contours.append(model_contours)

    def compute_mask(self):
        """
        Default is an empty array. Use the computed contours to fill the mask, not needed if the user wants an empty
        array.
        Returns
        -------

        """
        self.mask = np.zeros((self.dims[0], self.dims[1], self.dims[2]))
        if not self.empty_array:
            for ii, model in enumerate(self.models):

                model_contours = self.contours[ii]
                for jj, s in enumerate(self.slice_locations):
                    if len(model_contours[jj]) > 0:
                        frame = np.zeros((self.dims[1], self.dims[2]))
                        cv2.fillPoly(frame, np.array([model_contours[jj]], dtype=np.int32), 1)
                        self.mask[jj, :, :] = self.mask[jj, :, :] + frame

        self.mask = self.mask.astype(np.int8)

    def save_image(self, path):
        """
        Uses SimpleITK to write out the mask.

        Returns
        -------

        """
        image = sitk.GetImageFromArray(self.mask)
        image.SetSpacing(self.spacing)
        image.SetOrigin([self.bounds[0], self.bounds[2], self.bounds[4]])
        sitk.WriteImage(image, path)




from .conversion import ContourToDiscreteMesh, ContourToMask, ModelToMask
# from .creation import CreateDicomImage, CreateImageFromMask

from .mesh.volume import Volume
from .mesh.surface import Refinement

from .deformable.simpleitk import DeformableITK



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:
"""

import cv2
import numpy as np


def contours_from_mask(mask, plane='Axial'):
    mask = mask.astype(np.uint8)

    if plane == 'Axial':
        slices = mask.shape[0]
    elif plane == 'Coronal':
        slices = mask.shape[1]
    else:
        slices = mask.shape[2]

    contours = []
    for ii in range(slices):
        if plane == 'Axial':
            tuple_contour, _ = cv2.findContours(mask[ii, :, :], cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
            contours += [np.concatenate((np.vstack(t), ii * np.ones((len(t), 1))), axis=1) for t in tuple_contour]
        elif plane == 'Coronal':
            tuple_contour, _ = cv2.findContours(mask[:, ii, :], cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
            for t in tuple_contour:
                stack = np.vstack(t)
                contours += [np.vstack((stack[:, 0], ii * np.ones(len(t)), stack[:, 1])).T]
        else:
            tuple_contour, _ = cv2.findContours(mask[:, :, ii], cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
            contours += [np.concatenate((ii * np.ones((len(t), 1)), np.vstack(t)), axis=1) for t in tuple_contour]

    return contours



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:

Structure:

"""

import os
import copy

import numpy as np
import pandas as pd
import pyvista as pv
import SimpleITK as sitk

from scipy.spatial.transform import Rotation

import vtk
from vtkmodules.util import numpy_support


class Display(object):
    def __init__(self, dose):
        self.dose = dose

        self.matrix = copy.deepcopy(self.dose.matrix)
        self.spacing = copy.deepcopy(self.dose.spacing)
        self.origin = copy.deepcopy(self.dose.origin)

        self.slice_location = self.dose.compute_center(position=False, zyx=True)
        self.scroll_max = [self.dose.dimensions[0] - 1,
                           self.dose.dimensions[1] - 1,
                           self.dose.dimensions[2] - 1]
        self.secondary_array = None
        self.misc = {}

    def compute_matrix_pixel_to_position(self):
        matrix = copy.deepcopy(self.matrix)
        spacing = self.spacing

        pixel_to_position_matrix = np.identity(4, dtype=np.float32)
        pixel_to_position_matrix[:3, 0] = matrix[0, :] * spacing[0]
        pixel_to_position_matrix[:3, 1] = matrix[1, :] * spacing[1]
        pixel_to_position_matrix[:3, 2] = matrix[2, :] * spacing[2]
        pixel_to_position_matrix[:3, 3] = self.origin

        return pixel_to_position_matrix

    def compute_matrix_position_to_pixel(self):
        matrix = copy.deepcopy(self.matrix)
        spacing = self.spacing

        hold_matrix = np.identity(3, dtype=np.float32)
        hold_matrix[0, :] = matrix[0, :] / spacing[0]
        hold_matrix[1, :] = matrix[1, :] / spacing[1]
        hold_matrix[2, :] = matrix[2, :] / spacing[2]

        position_to_pixel_matrix = np.identity(4, dtype=np.float32)
        position_to_pixel_matrix[:3, :3] = hold_matrix
        position_to_pixel_matrix[:3, 3] = np.asarray(self.origin).dot(-hold_matrix.T)

        return position_to_pixel_matrix

    def compute_array(self, slice_plane):
        if self.secondary_array is None:
            if slice_plane == 'Axial':
                array = self.dose.array[self.slice_location[0], :, :]
            elif slice_plane == 'Coronal':
                array = self.dose.array[:, self.slice_location[1], :]
            else:
                array = self.dose.array[:, :, self.slice_location[2]]
        else:
            if slice_plane == 'Axial':
                array = self.secondary_array[self.slice_location[0], :, :]
            elif slice_plane == 'Coronal':
                array = self.secondary_array[:, self.slice_location[1], :]
            else:
                array = self.secondary_array[:, :, self.slice_location[2]]

        return array.astype(np.float32)

    def compute_index_positions(self, xyz):
        pixel_to_position_matrix = self.compute_matrix_pixel_to_position()
        location = np.asarray([xyz[0], xyz[1], xyz[2], 1])

        return location.dot(pixel_to_position_matrix.T)[:3]

    def compute_offaxis_array(self):
        loc = np.flip(self.slice_location)
        base_position_matrix = self.compute_matrix_pixel_to_position()
        slice_position = np.asarray([loc[0], loc[1], loc[2], 1]).dot(base_position_matrix.T)[:3]

        matrix_reshape = self.dose.matrix.reshape(1, 9)[0]
        vtk_image = vtk.vtkImageData()
        vtk_image.SetSpacing(self.dose.spacing)
        vtk_image.SetDirectionMatrix(matrix_reshape)
        vtk_image.SetDimensions(np.flip(self.dose.array.shape))
        vtk_image.SetOrigin(self.dose.origin)
        vtk_image.GetPointData().SetScalars(numpy_support.numpy_to_vtk(self.dose.array.flatten(order="C")))

        matrix = vtk.vtkMatrix4x4()
        for i in range(3):
            for j in range(3):
                matrix.SetElement(i, j, self.matrix[i, j])

        transform = vtk.vtkTransform()
        transform.SetMatrix(matrix)
        transform.Inverse()

        vtk_reslice = vtk.vtkImageReslice()
        vtk_reslice.SetInputData(vtk_image)
        vtk_reslice.SetResliceTransform(transform)
        vtk_reslice.SetInterpolationModeToLinear()
        vtk_reslice.SetOutputSpacing(self.dose.spacing)
        vtk_reslice.AutoCropOutputOn()
        vtk_reslice.SetBackgroundLevel(-3001)
        vtk_reslice.Update()

        reslice_data = vtk_reslice.GetOutput()
        new_origin = reslice_data.GetOrigin()
        self.origin = transform.TransformPoint(new_origin)
        dimensions = reslice_data.GetDimensions()

        position_to_pixel_matrix = self.compute_matrix_position_to_pixel()
        location = np.asarray([slice_position[0], slice_position[1], slice_position[2], 1])
        self.slice_location = list(np.flip(np.round(location.dot(position_to_pixel_matrix.T)[:3])).astype(np.int32))
        self.scroll_max = [dimensions[2] - 1, dimensions[1] - 1, dimensions[0] - 1]
        if self.slice_location[0] > dimensions[2] - 1:
            self.slice_location[0] = dimensions[2] - 1
        if self.slice_location[1] > dimensions[1] - 1:
            self.slice_location[1] = dimensions[1] - 1
        if self.slice_location[2] > dimensions[0] - 1:
            self.slice_location[2] = dimensions[0] - 1

        scalars = reslice_data.GetPointData().GetScalars()
        self.secondary_array = numpy_support.vtk_to_numpy(scalars).reshape(dimensions[2], dimensions[1], dimensions[0])

    def compute_scroll_max(self):
        if self.secondary_array is not None:
            self.scroll_max = [self.secondary_array.shape[0] - 1,
                               self.secondary_array.shape[1] - 1,
                               self.secondary_array.shape[2] - 1]
        else:
            self.scroll_max = [self.dose.dimensions[0] - 1,
                               self.dose.dimensions[1] - 1,
                               self.dose.dimensions[2] - 1]

    def compute_vtk_slice(self, slice_plane):
        matrix_reshape = np.linalg.inv(self.matrix).reshape(1, 9)[0]
        pixel_to_position_matrix = self.compute_matrix_pixel_to_position()
        if slice_plane == 'Axial':
            location = np.asarray([0, 0, self.slice_location[0], 1])
            if self.secondary_array is None:
                array_slice = self.dose.array[self.slice_location[0], :, :]
            else:
                array_slice = self.secondary_array[self.slice_location[0], :, :]
            array_shape = array_slice.shape
            dim = [array_shape[1], array_shape[0], 1]
        elif slice_plane == 'Coronal':
            location = np.asarray([0, self.slice_location[1], 0, 1])
            if self.secondary_array is None:
                array_slice = self.dose.array[:, self.slice_location[1], :]
            else:
                array_slice = self.secondary_array[:, self.slice_location[1], :]
            array_shape = array_slice.shape
            dim = [array_shape[1], 1, array_shape[0]]
        else:
            location = np.asarray([self.slice_location[2], 0, 0, 1])
            if self.secondary_array is None:
                array_slice = self.dose.array[:, :, self.slice_location[2]]
            else:
                array_slice = self.secondary_array[:, :, self.slice_location[2]]
            array_shape = array_slice.shape
            dim = [1, array_shape[1], array_shape[0]]

        slice_origin = location.dot(pixel_to_position_matrix.T)[:3]

        vtk_test = vtk.vtkImageData()
        vtk_test.SetSpacing(self.dose.spacing)
        vtk_test.SetDirectionMatrix(matrix_reshape)
        vtk_test.SetDimensions(dim)
        vtk_test.SetOrigin(slice_origin)
        vtk_test.GetPointData().SetScalars(numpy_support.numpy_to_vtk(array_slice.flatten(order="C")))

        return vtk_test

    def update_slice_location(self, scroll, slice_plane):
        if slice_plane == 'Axial':
            self.slice_location[0] = scroll
        elif slice_plane == 'Coronal':
            self.slice_location[1] = scroll
        else:
            self.slice_location[2] = scroll


class Dose(object):
    def __init__(self, dose):
        self.tags = dose.image_set
        self.array = dose.array

        self.dose_name = dose.dose_name
        self.modality = dose.modality

        self.patient_name = self.get_patient_name()
        self.mrn = self.get_mrn()
        self.birthdate = self.get_birthdate()
        self.date = self.get_date()
        self.time = self.get_time()
        self.series_uid = self.get_series_uid()
        self.acq_number = self.get_acq_number()
        self.frame_ref = self.get_frame_ref()
        self.window = self.get_window()

        self.filepaths = dose.filepaths
        self.sops = dose.sops

        self.plane = dose.plane
        self.spacing = dose.spacing
        self.dimensions = dose.dimensions
        self.orientation = dose.orientation
        self.origin = dose.origin
        self.matrix = dose.image_matrix

        self.camera_position = None
        self.misc = {}

        self.rois = {}
        self.display = Display(self)

    def get_patient_name(self):
        if 'PatientName' in self.tags[0]:
            return str(self.tags[0].PatientName).split('^')[:3]
        else:
            return 'missing'

    def get_mrn(self):
        if 'PatientID' in self.tags[0]:
            return str(self.tags[0].PatientID)
        else:
            return 'missing'

    def get_birthdate(self):
        if 'PatientBirthDate' in self.tags[0]:
            return str(self.tags[0].PatientBirthDate)
        else:
            return ''

    def get_date(self):
        if 'SeriesDate' in self.tags[0]:
            return self.tags[0].SeriesDate
        elif 'ContentDate' in self.tags[0]:
            return self.tags[0].ContentDate
        elif 'AcquisitionDate' in self.tags[0]:
            return self.tags[0].AcquisitionDate
        elif 'StudyDate' in self.tags[0]:
            return self.tags[0].StudyDate
        else:
            return '00000'

    def get_time(self):
        if 'SeriesTime' in self.tags[0]:
            return self.tags[0].SeriesTime
        elif 'ContentTime' in self.tags[0]:
            return self.tags[0].ContentTime
        elif 'AcquisitionTime' in self.tags[0]:
            return self.tags[0].AcquisitionTime
        elif 'StudyTime' in self.tags[0]:
            return self.tags[0].StudyTime
        else:
            return '00000'

    def get_study_uid(self):
        if 'StudyInstanceUID' in self.tags[0]:
            return self.tags[0].StudyInstanceUID
        else:
            return '00000.00000'

    def get_series_uid(self):
        if 'SeriesInstanceUID' in self.tags[0]:
            return self.tags[0].SeriesInstanceUID
        else:
            return '00000.00000'

    def get_acq_number(self):
        if 'AcquisitionNumber' in self.tags[0]:
            return self.tags[0].AcquisitionNumber
        else:
            return '1'

    def get_frame_ref(self):
        if 'FrameOfReferenceUID' in self.tags[0]:
            return self.tags[0].FrameOfReferenceUID
        else:
            return '00000.00000'

    def get_window(self):
        if (0x0028, 0x1050) in self.tags[0] and (0x0028, 0x1051) in self.tags[0]:
            center = self.tags[0].WindowCenter
            width = self.tags[0].WindowWidth

            if not isinstance(center, float):
                center = center[0]

            if not isinstance(width, float):
                width = width[0]

            return [int(center) - int(np.round(width / 2)), int(center) + int(np.round(width / 2))]

        elif self.array is not None:
            return [np.min(self.array), np.max(self.array)]

        else:
            return [0, 1]

    def get_specific_tag(self, tag):
        if tag in self.tags[0]:
            return self.tags[0][tag]
        else:
            return None

    def compute_aspect(self, slice_plane):
        if slice_plane == 'Axial':
            aspect = np.round(self.spacing[0] / self.spacing[1], 2)
        elif slice_plane == 'Coronal':
            aspect = np.round(self.spacing[0] / self.spacing[2], 2)
        else:
            aspect = np.round(self.spacing[1] / self.spacing[2], 2)

        return aspect

    def compute_bounds(self):
        shape = self.array.shape
        matrix_reshape = self.matrix.reshape(1, 9)[0]
        vtk_image = vtk.vtkImageData()
        vtk_image.SetSpacing(self.spacing)
        vtk_image.SetDirectionMatrix(matrix_reshape)
        vtk_image.SetDimensions([shape[1], shape[2], shape[0]])
        vtk_image.SetOrigin(self.origin)

        x_min, x_max, y_min, y_max, z_min, z_max = vtk_image.GetBounds()

        return [x_min, x_max, y_min, y_max, z_min, z_max]

    def compute_center(self, position=True, zyx=False):
        pixel_index = [int(self.dimensions[2] / 2),
                       int(self.dimensions[1] / 2),
                       int(self.dimensions[0] / 2)]

        if position:
            pixel_to_position_matrix = self.display.compute_matrix_pixel_to_position()
            location = np.asarray([pixel_index[0], pixel_index[1], pixel_index[2], 1])

            center = location.dot(pixel_to_position_matrix.T)[:3]
            if zyx:
                return np.flip(center)
            else:
                return center

        else:
            if zyx:
                return [pixel_index[2], pixel_index[1], pixel_index[0]]
            else:
                return pixel_index

    def compute_corner_positions(self):
        shape = self.array.shape
        matrix_reshape = self.matrix.reshape(1, 9)[0]
        vtk_image = vtk.vtkImageData()
        vtk_image.SetSpacing(self.spacing)
        vtk_image.SetDirectionMatrix(matrix_reshape)
        vtk_image.SetDimensions([shape[1], shape[2], shape[0]])
        vtk_image.SetOrigin(self.origin)

        x_min, x_max, y_min, y_max, z_min, z_max = vtk_image.GetBounds()

        corner_points = [(x_min, y_min, z_min),
                         (x_max, y_min, z_min),
                         (x_max, y_max, z_min),
                         (x_min, y_max, z_min),
                         (x_min, y_min, z_max),
                         (x_max, y_min, z_max),
                         (x_max, y_max, z_max),
                         (x_min, y_max, z_max)]

        return corner_points

    def compute_corner_sides(self):
        corner_points = self.compute_corner_positions()
        points = [corner_points[0], corner_points[4], corner_points[7], corner_points[3],
                  corner_points[1], corner_points[2], corner_points[6], corner_points[5]]
        faces = [4, 0, 1, 2, 3,
                 4, 4, 5, 6, 7,
                 4, 0, 4, 7, 1,
                 4, 3, 2, 6, 5,
                 4, 0, 3, 5, 4,
                 4, 1, 7, 6, 2]

        return pv.PolyData(points, faces)

    def compute_pixel(self, position):
        matrix = copy.deepcopy(self.matrix)

        hold_matrix = np.identity(3, dtype=np.float32)
        hold_matrix[0, :] = matrix[0, :] / self.spacing[0]
        hold_matrix[1, :] = matrix[1, :] / self.spacing[1]
        hold_matrix[2, :] = matrix[2, :] / self.spacing[2]

        position_to_pixel_matrix = np.identity(4, dtype=np.float32)
        position_to_pixel_matrix[:3, :3] = hold_matrix
        position_to_pixel_matrix[:3, 3] = np.asarray(self.origin).dot(-hold_matrix.T)

        location = np.asarray([position[0], position[1], position[2], 1])

        return (np.round(location.dot(position_to_pixel_matrix.T)[:3])).astype(np.int32)

    def compute_position(self, xyz):
        matrix = copy.deepcopy(self.matrix)

        pixel_to_position_matrix = np.identity(4, dtype=np.float32)
        pixel_to_position_matrix[:3, 0] = matrix[0, :] * self.spacing[0]
        pixel_to_position_matrix[:3, 1] = matrix[1, :] * self.spacing[1]
        pixel_to_position_matrix[:3, 2] = matrix[2, :] * self.spacing[2]
        pixel_to_position_matrix[:3, 3] = self.origin

        pixel_to_position_matrix = self.compute_matrix_pixel_to_position()
        location = np.asarray([xyz[0], xyz[1], xyz[2], 1])

        return location.dot(pixel_to_position_matrix.T)[:3]

    def compute_matrix_pixel_to_position(self):
        matrix = copy.deepcopy(self.matrix)
        spacing = self.spacing

        pixel_to_position_matrix = np.identity(4, dtype=np.float32)
        pixel_to_position_matrix[:3, 0] = matrix[0, :] * spacing[0]
        pixel_to_position_matrix[:3, 1] = matrix[1, :] * spacing[1]
        pixel_to_position_matrix[:3, 2] = matrix[2, :] * spacing[2]
        pixel_to_position_matrix[:3, 3] = self.origin

        return pixel_to_position_matrix

    def reset_array(self):
        self.display.secondary_array = None
        self.display.matrix = copy.deepcopy(self.matrix)
        self.display.origin = copy.deepcopy(self.origin)
        self.display.slice_location = self.compute_center(position=False, zyx=True)

    def retrieve_angles(self, order='ZXY'):
        rotation = Rotation.from_matrix(self.display.matrix[:3, :3])

        return rotation.as_euler(order, degrees=True)

    def retrieve_array_plane(self, slice_plane):
        return self.display.compute_array(slice_plane=slice_plane)

    def retrieve_slice_location(self, slice_plane):
        if slice_plane == 'Axial':
            return self.display.slice_location[0]

        elif slice_plane == 'Coronal':
            return self.display.slice_location[1]

        else:
            return self.display.slice_location[2]

    def retrieve_slice_position(self, slice_plane=None):
        pixel_to_position_matrix = self.display.compute_matrix_pixel_to_position()

        if slice_plane is None:
            location = np.asarray([self.display.slice_location[2],
                                   self.display.slice_location[1],
                                   self.display.slice_location[0], 1])
        else:
            if slice_plane == 'Axial':
                location = np.asarray([0, 0, self.display.slice_location[0], 1])
            elif slice_plane == 'Coronal':
                location = np.asarray([0, self.display.slice_location[1], 0, 1])
                print(location)
            else:
                location = np.asarray([self.display.slice_location[2], 0, 0, 1])

        return location.dot(pixel_to_position_matrix.T)[:3]

    def retrieve_scroll_max(self, slice_plane):
        if slice_plane == 'Axial':
            return self.display.scroll_max[0]

        elif slice_plane == 'Coronal':
            return self.display.scroll_max[1]

        else:
            return self.display.scroll_max[2]

    def retrieve_vtk_slice(self, slice_plane):
        return self.display.compute_vtk_slice(slice_plane)

    def retrieve_vtk_volume(self, slice_plane):
        return self.display.compute_vtk_volume(slice_plane)

    def save_image(self, path):
        variable_names = self.__dict__.keys()
        column_names = [name for name in variable_names if name not in ['tags', 'array', 'display', 'rois']]

        df = pd.DataFrame(index=[0], columns=column_names)
        for name in column_names:
            df.at[0, name] = getattr(self, name)

        df.to_pickle(os.path.join(path, 'info.p'))
        np.save(os.path.join(path, 'tags.npy'), self.tags, allow_pickle=True)
        np.save(os.path.join(path, 'array.npy'), self.array, allow_pickle=True)

    def update_rotation(self, r_x=0, r_y=0, r_z=0, base=True):
        if r_x != 0 or r_y != 0 or r_z != 0:
            r = Rotation.from_euler('xyz', [r_x, r_y, r_z], degrees=True)
            new_matrix = r.as_matrix()

            if base:
                base_matrix = copy.deepcopy(self.matrix)
                self.display.matrix = new_matrix @ base_matrix
            else:
                self.display.matrix = new_matrix @ self.display.matrix

            self.display.compute_offaxis_array()
            self.display.compute_scroll_max()
        else:
            self.display.compute_scroll_max()
            self.reset_array()



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org


Description:

Functions:

"""

import os

import zipfile
from PIL import ImageColor
import xml.etree.ElementTree as ET

import numpy as np
import pyvista as pv

from ..structure.image import Image
from ..utils.creation import CreateImageFromMask
from ..utils.conversion import ModelToMask

from ..data import Data


class ThreeMfReader(object):
    """
    Converts 3mf file to pyvista polydata mesh.
    """
    def __init__(self, file, roi_name=None):
        self.file = file
        self.roi_name = roi_name

    def load(self):
        """
        Loads in the 3mf file, gets the vertices/vertice colors/triangles and creates a polydata 3D model using pyvista.

        Returns
        -------

        """
        if self.roi_name is not None:
            roi_name = self.roi_name
        else:
            roi_name = self.file.split('/')[-1].split('.3mf')[0]

        namespace = {"3mf": "http://schemas.microsoft.com/3dmanufacturing/core/2015/02",
                     "m": "http://schemas.microsoft.com/3dmanufacturing/material/2015/02"}

        archive = zipfile.ZipFile(self.file, "r")
        root = ET.parse(archive.open("3D/3dmodel.model"))
        color_list = list()
        colors = root.findall('.//m:color', namespace)
        if colors:
            for color in colors:
                color_list.append(color.get("color", 0))

        obj = root.findall("./3mf:resources/3mf:object", namespace)[0]
        triangles = obj.findall(".//3mf:triangle", namespace)

        vertex_list = []
        for vertex in obj.findall(".//3mf:vertex", namespace):
            vertex_list.append([vertex.get("x"), vertex.get("y"), vertex.get("z")])

        triangle_list = np.empty((1, 4 * len(triangles)))
        vertices_color = np.zeros((len(vertex_list), 3))
        for ii, triangle in enumerate(triangles):
            v1 = int(triangle.get("v1"))
            v2 = int(triangle.get("v2"))
            v3 = int(triangle.get("v3"))
            tricolor = self.color_avg(color_list, (triangle.get("p1")), (triangle.get("p2")), (triangle.get("p3")))
            rgb_color = list(ImageColor.getcolor(tricolor, "RGB")[0:3])
            vertices_color[v1] = rgb_color
            vertices_color[v2] = rgb_color
            vertices_color[v3] = rgb_color
            triangle_list[0, ii * 4:(ii + 1) * 4] = [3, v1, v2, v3]

        mesh = pv.PolyData(np.float64(np.asarray(vertex_list)), triangle_list[0, :].astype(int))
        mesh['colors'] = np.abs(255-vertices_color)

        decimate_mesh = mesh.decimate_pro(1 - (50000 / len(mesh.points)))

        image_name = 'CT ' + '0' + str(len(Data.image_list) + 1)

        model_to_mask = ModelToMask([decimate_mesh])
        mask = model_to_mask.mask

        new_image = CreateImageFromMask(mask, model_to_mask.origin, model_to_mask.spacing, image_name)
        Data.image[image_name] = Image(new_image)
        Data.image_list += [image_name]

        Data.image[image_name].create_roi(name=roi_name, visible=False, filepath=self.file)
        Data.image[image_name].rois[roi_name].add_mesh(decimate_mesh)
        Data.image[image_name].rois[roi_name].color = [128, 128, 128]
        Data.image[image_name].rois[roi_name].multi_color = True
        Data.match_rois()

    @staticmethod
    def color_avg(color_list, p1, p2, p3):
        """
        Get the average color from color list.

        Parameters
        ----------
        color_list
        p1
        p2
        p3

        Returns
        -------

        """
        p2rgb = None
        p3rgb = None

        p1hex = color_list[int(p1)]
        value = p1hex.lstrip('#')
        lv = len(value)
        p1rgb = tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))

        if isinstance(p2, int):
            p2hex = color_list[int(p2)]
            value = p2hex.lstrip('#')
            lv = len(value)
            p2rgb = tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))

        if isinstance(p3, int):
            p3hex = color_list[int(p3)]
            value = p3hex.lstrip('#')
            lv = len(value)
            p3rgb = tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))

        if p2rgb is not None and p3rgb is not None:
            rgbAvg = np.average(np.array(p1rgb), np.array(p2rgb), np.array(p3rgb))

        elif p2rgb is not None:
            rgbAvg = np.average(np.array(p1rgb), np.array(p2rgb))

        else:
            rgbAvg = p1rgb

        hexAvg = '#%02x%02x%02x' % (rgbAvg[0], rgbAvg[1], rgbAvg[2])

        return hexAvg





"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org


Description:

Functions:

"""

import os
import copy
import time
import gdcm
import threading
from struct import unpack

import numpy as np
import pandas as pd
import pydicom as dicom
from openpyxl.descriptors import NoneSet
from pydicom.uid import generate_uid

from ..structure.deformable import Deformable
from ..structure.dose import Dose
from ..structure.image import Image
from ..structure.rigid import Rigid

from ..data import Data


def sort_images_by_datetime():
    """
    Reorders the global Data.image dictionary and Data.image_list based on the DICOM image acquisition date and time.
    """
    date_time = [str(Data.image[name].date) + str(Data.image[name].time) for name in Data.image_list]
    new_key_order = [Data.image_list[idx] for idx in np.argsort(date_time)]

    Data.image = {key: Data.image[key] for key in new_key_order}
    Data.image_list = list(Data.image.keys())


def thread_process_dicom(path, stop_before_pixels=False):
    """
    Reads a DICOM file from the given path, optionally stopping before loading pixel data.
    """
    try:
        datasets = dicom.dcmread(str(path), stop_before_pixels=stop_before_pixels)
    except:
        datasets = []

    return datasets


class DicomReader(object):
    """
    Main class to read, organize, and process a collection of DICOM files.
    - Multithreaded reading of DICOMs
    - Sorting by modality and orientation
    - Image creation and registration
    - Associating RTSTRUCT/RTDOSE data with corresponding images
    """
    def __init__(self, files, only_tags, only_modality, only_load_roi_names, clear):
        """
        Initializes DICOM reading parameters.

        Arguments:
            files: dictionary containing DICOM file paths.
            only_tags: if True, reads only metadata (no pixel arrays).
            only_modality: list of modalities to process (CT, MR, etc.).
            only_load_roi_names: if True, loads only ROI names (not contours).
            clear: if True, clears the global Data structure before loading.
        """
        self.files = files
        self.only_tags = only_tags
        self.only_modality = only_modality
        self.only_load_roi_names = only_load_roi_names

        if clear:
            Data.clear()

        self.ds = []
        self.ds_modality = {key: [] for key in ['CT', 'MR', 'PT', 'US', 'DX', 'RF', 'CR', 'RTSTRUCT', 'REG', 'RTDOSE']}

    def load(self, display_time=False):
        """
        Main entry point to read and organize all DICOM data.
        """
        t1 = time.time()
        self.read()
        self.separate_modalities_and_images()
        self.image_creation()
        sort_images_by_datetime()
        t2 = time.time()

        if display_time:
            print('Dicom Read Time: ', t2 - t1)

    def read(self):
        """
        Reads all DICOM files in parallel threads.
        """
        threads = []

        def read_file_thread(file_path):
            self.ds.append(thread_process_dicom(file_path, stop_before_pixels=self.only_tags))

        for file_path in self.files['Dicom']:
            thread = threading.Thread(target=read_file_thread, args=(file_path,))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

    def separate_modalities_and_images(self):
        """
        Purpose:
        Sorts all loaded DICOM datasets (self.ds) into groups by modality, series, orientation, and slice position.

        Divides DICOMs by Modality (e.g., CT, MR, RTSTRUCT).
            For 3D modalities (CT, MR, PT):
                Groups by SeriesInstanceUID and AcquisitionNumber.
                Sorts slices by spatial position and image orientation.
                Determines the acquisition plane (Axial, Coronal, Sagittal).
            For 2D or non-image modalities (US, CR, RTSTRUCT, etc.):
                Simply groups without sorting.

        Populates self.ds_modality[modality] with properly ordered image sets.
        """
        for modality in list(self.ds_modality.keys()):
            images_in_modality = [d for d in self.ds if (0x0008, 0x0060) in d and d['Modality'].value == modality]
            if len(images_in_modality) > 0 and modality in self.only_modality:
                if modality in ['US', 'DX', 'RF', 'CR', 'RTSTRUCT', 'REG', 'RTDOSE']:
                    for image in images_in_modality:
                        self.ds_modality[modality] += [image]

                else:
                    sorting_tags = []
                    for img in images_in_modality:
                        orient = np.asarray(img['ImageOrientationPatient'].value)
                        pos = np.asarray(img['ImagePositionPatient'].value)
                        if 'AcquisitionNumber' in img and img['AcquisitionNumber'].value is not None:
                            acq = np.int64(img['AcquisitionNumber'].value)
                        else:
                            acq = 1

                        sorting_tags += [[img['SeriesInstanceUID'].value, acq, orient[0], orient[1], orient[2],
                                          orient[3], orient[4], orient[5], pos[0], pos[1], pos[2]]]
                    sorting_tags = np.asarray(sorting_tags)
                    unique_series = np.unique(np.asarray(sorting_tags[:, 0]), axis=0)
                    for series in unique_series:
                        idx = np.where(sorting_tags[:, 0] == series)
                        series_tags = sorting_tags[idx[0], :]
                        series_image = [images_in_modality[ii] for ii in idx[0]]

                        orientations = series_tags[:, 2:8].astype(np.float64)
                        _, indices = np.unique(np.round(orientations, 3), axis=0, return_index=True)
                        unique_orientations = [orientations[ind].astype(np.float64) for ind in indices]
                        for orient in unique_orientations:
                            orient_idx = np.where((np.round(orientations[:, 0], 3) == np.round(orient[0], 3)) &
                                                  (np.round(orientations[:, 1], 3) == np.round(orient[1], 3)) &
                                                  (np.round(orientations[:, 2], 3) == np.round(orient[2], 3)) &
                                                  (np.round(orientations[:, 3], 3) == np.round(orient[3], 3)) &
                                                  (np.round(orientations[:, 4], 3) == np.round(orient[4], 3)) &
                                                  (np.round(orientations[:, 5], 3) == np.round(orient[5], 3)))

                            orient_tags = np.asarray([series_tags[orient] for orient in orient_idx[0]])
                            orient_image = [series_image[orient] for orient in orient_idx[0]]
                            correct_orientation = orient_tags[0, 2:8].astype(np.float64)

                            x = np.abs(correct_orientation[0]) + np.abs(correct_orientation[3])
                            y = np.abs(correct_orientation[1]) + np.abs(correct_orientation[4])
                            z = np.abs(correct_orientation[2]) + np.abs(correct_orientation[5])

                            row_direction = correct_orientation[:3]
                            column_direction = correct_orientation[3:]
                            slice_direction = np.cross(row_direction, column_direction)

                            unique_acq = np.unique(orient_tags[:, 1])

                            acq_plane = []
                            acq_images = []
                            acq_positions = []
                            for acq in unique_acq:
                                orient_idx = np.where(orient_tags == acq)[0]
                                acq_tags = orient_tags[orient_idx]
                                acq_image = [orient_image[ii] for ii in orient_idx]
                                position_tags = np.asarray([np.asarray(t[8:]).astype(np.double) for t in acq_tags])

                                if x < y and x < z:
                                    acq_plane += ['Sagittal']
                                    if slice_direction[0] > 0:
                                        slice_idx = np.argsort(position_tags[:, 0])
                                    else:
                                        slice_idx = np.argsort(position_tags[:, 0])[::-1]
                                elif y < x and y < z:
                                    acq_plane += ['Coronal']
                                    if slice_direction[1] > 0:
                                        slice_idx = np.argsort(position_tags[:, 1])
                                    else:
                                        slice_idx = np.argsort(position_tags[:, 1])[::-1]
                                else:
                                    acq_plane += ['Axial']
                                    if slice_direction[2] > 0:
                                        slice_idx = np.argsort(position_tags[:, 2])
                                    else:
                                        slice_idx = np.argsort(position_tags[:, 2])[::-1]

                                acq_images += [np.asarray([acq_image[idx] for idx in slice_idx])]
                                acq_positions += [np.asarray([acq_tags[idx] for idx in slice_idx])]

                            if len(acq_positions) > 1:
                                exclude_images = np.zeros((len(acq_positions), 1))
                                for ii in range(len(acq_positions)):
                                    for jj in range(len(acq_positions)):
                                        if ii != jj:
                                            if acq_plane[0] == 'Sagittal':
                                                base_first = acq_positions[ii][0, 8]
                                                base_last = acq_positions[ii][-1, 8]
                                                check_first = acq_positions[jj][0, 8]
                                                check_last = acq_positions[jj][-1, 8]
                                            elif acq_plane[0] == 'Coronal':
                                                base_first = acq_positions[ii][0, 9]
                                                base_last = acq_positions[ii][-1, 9]
                                                check_first = acq_positions[jj][0, 9]
                                                check_last = acq_positions[jj][-1, 9]
                                            else:
                                                base_first = acq_positions[ii][0, 10]
                                                base_last = acq_positions[ii][-1, 10]
                                                check_first = acq_positions[jj][0, 10]
                                                check_last = acq_positions[jj][-1, 10]

                                            base_first = np.float64(base_first)
                                            base_last = np.float64(base_last)
                                            check_first = np.float64(check_first)
                                            check_last = np.float64(check_last)

                                            if base_first > check_first and base_first > check_last:
                                                pass

                                            elif base_last < check_first and base_last < check_last:
                                                pass

                                            else:
                                                exclude_images[ii] = 1

                                if np.sum(exclude_images) == 0:
                                    if acq_plane[0] == 'Sagittal':
                                        pos = np.asarray([[p[0, 8], p[-1, 8]] for p in acq_positions])
                                    elif acq_plane[0] == 'Coronal':
                                        pos = np.asarray([[p[0, 9], p[-1, 9]] for p in acq_positions])
                                    else:
                                        pos = np.asarray([[p[0, 10], p[-1, 10]] for p in acq_positions]).astype(
                                            np.float64)

                                    pos_idx = np.argsort(pos[:, 0])
                                    pos_sort = pos[pos_idx]
                                    pos_diff = [pos_sort[ii + 1, 0] - pos_sort[ii, 1] for ii in range(len(pos) - 1)]
                                    if len(np.unique(np.round(pos_diff, 2))) == 1:
                                        img = []
                                        for ii in pos_idx:
                                            for acq in acq_images[ii]:
                                                img += [acq]
                                        self.ds_modality[modality] += [img]

                                    else:
                                        for img in acq_images:
                                            self.ds_modality[modality] += [img.tolist()]

                                else:
                                    for img in acq_images:
                                        self.ds_modality[modality] += [img.tolist()]

                            else:
                                for img in acq_images:
                                    self.ds_modality[modality] += [img.tolist()]

    def image_creation(self):
        """
        Converts grouped DICOM datasets into actual image/structure objects and integrates them into the global Data structure.

        Image Modalities:
            Uses specialized readers (Read3D, ReadXRay, ReadRF, ReadUS) for CT/MR/PT, DX/CR, RF, US images.

        RTSTRUCT (contours):
            Reads using ReadRTStruct, associates to matching image via Data.image[...].
            Calls Data.match_rois() and Data.match_pois().

        Registration and Dose:
            Reads REG and RTDOSE modalities using their specific readers (ReadREG, ReadRTDose).
        """
        for modality in ['CT', 'MR', 'PT', 'DX', 'RF', 'CR', 'US']:
            for image_set in self.ds_modality[modality]:
                if modality in ['CT', 'MR', 'PT']:
                    Read3D(image_set, self.only_tags)

                elif modality in ['DX', 'CR']:
                    ReadXRay(image_set, self.only_tags)

                elif modality == 'RF':
                    ReadRF(image_set, self.only_tags)

                elif modality == 'US':
                    ReadUS(image_set, self.only_tags)

        for modality in ['RTSTRUCT']:
            for image_set in self.ds_modality[modality]:
                read_rtstruct = ReadRTStruct(image_set, self.only_tags)
                if read_rtstruct.match_image_name is not None:
                    Data.image[read_rtstruct.match_image_name].input_rtstruct(read_rtstruct)
                else:
                    print('dicom: rtstruct has no matching image')

            Data.match_rois()
            Data.match_pois()

        for modality in ['REG']:
            for image_set in self.ds_modality[modality]:
                ReadREG(image_set, self.only_tags)

        for modality in ['RTDOSE']:
            for image_set in self.ds_modality[modality]:
                ReadRTDose(image_set, self.only_tags)


class Read3D(object):
    """
    Handles reading and constructing 3D medical image volumes (specifically CT/MR/PT modalities) from a list of DICOM
    slices. It extracts geometry, orientation, and pixel data to create a volumetric representation of the image and
    registers it in the global Data structure.
    """

    def __init__(self, image_set, only_tags):
        """
        image_set: list of DICOM slices (or a single slice).
        only_tags: if True, loads only DICOM header data (no pixel array).

        Creates an Image object and adds it to Data.image and Data.image_list.
        """
        if isinstance(image_set, list):
            self.image_set = image_set
        else:
            self.image_set = [image_set]
        self.only_tags = only_tags

        self.unverified = None
        self.base_position = None
        self.skipped_slice = None
        self.sections = None
        self.rgb = False

        self.modality = self.image_set[0].Modality
        self.filepaths = [image.filename for image in self.image_set]
        self.sops = [image.SOPInstanceUID for image in self.image_set]

        self.array = None
        if not self.only_tags:
            self._compute_array()

        self.orientation = self._compute_orientation()
        self.plane = self._compute_plane()
        self.spacing = self._compute_spacing()
        self.dimensions = self._compute_dimensions()
        self._verify_axial_orientation()

        self.image_matrix = self._compute_image_matrix()
        self.image_name = create_image_name(self.modality)

        image = Image(self)
        Data.image[self.image_name] = image
        Data.image_list += [self.image_name]

    def _compute_array(self):
        """
        Applies rescale slope/intercept for intensity normalization, then stacks slices. Deletes PixelData from each
        DICOM object to free memory.
        """

        image_slices = []
        for _slice in self.image_set:
            if (0x0028, 0x1052) in _slice:
                intercept = _slice.RescaleIntercept
            else:
                intercept = 0

            if (0x0028, 0x1053) in _slice:
                slope = _slice.RescaleSlope
            else:
                slope = 1

            image_slices.append(((_slice.pixel_array * slope) + intercept).astype('int16'))

            del _slice.PixelData

        self.array = np.asarray(image_slices)

    def _compute_orientation(self):
        """
        Extracts the image’s directional cosines (row/column orientation).
        """
        orientation = np.asarray([1, 0, 0, 0, 1, 0])
        if 'ImageOrientationPatient' in self.image_set[0]:
            orientation = np.asarray(self.image_set[0]['ImageOrientationPatient'].value)

        else:
            if 'SharedFunctionalGroupsSequence' in self.image_set[0]:
                seq_str = 'SharedFunctionalGroupsSequence'
                if 'PlaneOrientationSequence' in self.image_set[0][0][seq_str][0]:
                    plane_str = 'PlaneOrientationSequence'
                    image_str = 'ImageOrientationPatient'
                    orientation = np.asarray(self.image_set[0][0][seq_str][0][plane_str][0][image_str].value)

                else:
                    self.unverified = 'Orientation'

            else:
                self.unverified = 'Orientation'

        return orientation

    def _compute_plane(self):
        """
        Determines the anatomical plane (Axial, Coronal, or Sagittal) based on the orientation vectors.
        """
        x = np.abs(self.orientation[0]) + np.abs(self.orientation[3])
        y = np.abs(self.orientation[1]) + np.abs(self.orientation[4])
        z = np.abs(self.orientation[2]) + np.abs(self.orientation[5])

        if x < y and x < z:
            return 'Sagittal'
        elif y < x and y < z:
            return 'Coronal'
        else:
            return 'Axial'

    def _compute_spacing(self):
        """
        Creates 3 axis spacing by inplane pixel spacing the slice thickness
        """
        inplane_spacing = [1, 1]
        slice_thickness = np.double(self.image_set[0].SliceThickness)

        if 'PixelSpacing' in self.image_set[0]:
            inplane_spacing = self.image_set[0].PixelSpacing

        elif 'ContributingSourcesSequence' in self.image_set[0]:
            sequence = 'ContributingSourcesSequence'
            if 'DetectorElementSpacing' in self.image_set[0][sequence][0]:
                inplane_spacing = self.image_set[0][sequence][0]['DetectorElementSpacing']

        elif 'PerFrameFunctionalGroupsSequence' in self.image_set[0]:
            sequence = 'PerFrameFunctionalGroupsSequence'
            if 'PixelMeasuresSequence' in self.image_set[0][sequence][0]:
                inplane_spacing = self.image_set[0][sequence][0]['PixelMeasuresSequence'][0]['PixelSpacing']

        if len(self.image_set) > 1:
            row_direction = self.orientation[:3]
            column_direction = self.orientation[3:]
            slice_direction = np.cross(row_direction, column_direction)

            first = np.dot(slice_direction, self.image_set[0].ImagePositionPatient)
            second = np.dot(slice_direction, self.image_set[1].ImagePositionPatient)
            last = np.dot(slice_direction, self.image_set[-1].ImagePositionPatient)
            first_last_spacing = np.asarray((last - first) / (len(self.image_set) - 1))
            if np.abs((second - first) - first_last_spacing) > 0.01:
                if not self.only_tags:
                    self._find_skipped_slices(slice_direction)
                slice_thickness = second - first
            else:
                slice_thickness = np.asarray((last - first) / (len(self.image_set) - 1))

        if self.plane == 'Axial':
            return np.asarray([inplane_spacing[1], inplane_spacing[0], slice_thickness])

        elif self.plane == 'Coronal':
            return np.asarray([inplane_spacing[1], slice_thickness, inplane_spacing[0]])

        else:
            return np.asarray([slice_thickness, inplane_spacing[1], inplane_spacing[0]])

    def _compute_dimensions(self):
        """
        Creates dimensions by columns, rows, and number of slices.
        """
        shape = self.array.shape
        if self.plane == 'Axial':
            return np.asarray([shape[0], shape[1], shape[2]])

        elif self.plane == 'Coronal':
            return np.asarray([shape[1], shape[0], shape[2]])

        else:
            return np.asarray([shape[1], shape[2], shape[0]])

    def _compute_image_matrix(self):
        """
        Computes the image matrix using the image orientation.
        """
        row_direction = self.orientation[:3]
        column_direction = self.orientation[3:]
        slice_direction = np.cross(row_direction, column_direction)

        mat = np.identity(3, dtype=np.float32)
        mat[0, :3] = row_direction
        mat[1, :3] = column_direction
        mat[2, :3] = slice_direction

        return mat

    def _verify_axial_orientation(self):
        """
        Ensures that the 3D volume is oriented correctly in physical space.
            - Computes all 8 corner coordinates of the image volume using orientation and spacing.
            - Identifies the physical origin and corrects rotations or flips via np.rot90 and transpositions.
            - Adjusts orientation vectors based on corrected axes.
        """
        shape = self.array.shape
        if self.plane == 'Axial':
            spacing = self.spacing
        elif self.plane == 'Coronal':
            spacing = [self.spacing[0], self.spacing[2], self.spacing[1]]
        else:
            spacing = [self.spacing[1], self.spacing[2], self.spacing[0]]

        slices = shape[0] - 1
        y = shape[1] - 1
        x = shape[2] - 1

        origin = np.asarray(self.image_set[0]['ImagePositionPatient'].value)

        row_direction = self.orientation[:3]
        column_direction = self.orientation[3:]
        slice_direction = np.cross(row_direction, column_direction)

        corners = np.zeros((8, 3))
        corners[0] = origin
        corners[1] = origin + (x * spacing[0] * row_direction)
        corners[2] = origin + (y * spacing[1] * column_direction)
        corners[3] = (origin + (x * spacing[0] * row_direction) + (y * spacing[1] * column_direction))

        corners[4] = origin + (slices * spacing[2] * slice_direction)
        corners[5] = (origin + (slices * spacing[2] * slice_direction) + (x * spacing[0] * row_direction))
        corners[6] = (origin + (slices * spacing[2] * slice_direction) + (y * spacing[1] * column_direction))
        corners[7] = (origin + (slices * spacing[2] * slice_direction) + (x * spacing[0] * row_direction) +
                      (y * spacing[1] * column_direction))

        corner_idx = np.argmin(np.sum(corners, axis=1))
        if corner_idx != 0:
            self.origin = corners[corner_idx]
            if self.plane == "Axial":
                if corner_idx == 1:
                    self.array = np.rot90(self.array, 1, (1, 2))
                elif corner_idx == 2:
                    self.array = np.rot90(self.array, 3, (1, 2))
                else:
                    self.array = np.rot90(self.array, 2, (1, 2))

                if corner_idx < 4:
                    square = corners[:4, :]
                else:
                    square = corners[4:, :]

            elif self.plane == 'Coronal':
                self.array = np.rot90(self.array, 1, (0, 1))

                s1 = np.argsort(corners[:4, 2])
                s2 = np.argsort(corners[4:, 2]) + 4

                square = [corners[s1[0]], corners[s1[1]], corners[s2[0]], corners[s2[1]]]

            else:
                self.array = np.flip(np.rot90(self.array, 1, (0, 1)).transpose(0, 2, 1), axis=2)

                s1 = np.argsort(corners[:4, 2])
                s2 = np.argsort(corners[4:, 2]) + 4

                square = [corners[s1[0]], corners[s1[1]], corners[s2[0]], corners[s2[1]]]

            distances = np.asarray([np.linalg.norm(corners[corner_idx, :] - s) for s in square])
            sorted_args = np.argsort(distances)

            c1 = square[sorted_args[1]] - corners[corner_idx]
            c2 = square[sorted_args[2]] - corners[corner_idx]

            if np.abs(c1[0]) > np.abs(c2[0]):
                # self.orientation[:3] = c1 / self.spacing / np.flip(self.dimensions - 1)
                # self.orientation[3:] = c2 / self.spacing / np.flip(self.dimensions - 1)
                # self.orientation[:3] = c1 * self.spacing / np.linalg.norm(c1 * self.spacing)
                # self.orientation[3:] = c2 * self.spacing / np.linalg.norm(c2 * self.spacing)
                self.orientation[:3] = c1 / (self.spacing[0] * self.dimensions[2])
                self.orientation[3:] = c2 / (self.spacing[1] * self.dimensions[1])
            else:
                # self.orientation[:3] = c2 / self.spacing / np.flip(self.dimensions - 1)
                # self.orientation[3:] = c1 / self.spacing / np.flip(self.dimensions - 1)
                # self.orientation[:3] = c2 * self.spacing / np.linalg.norm(c2 * self.spacing)
                # self.orientation[3:] = c1 * self.spacing / np.linalg.norm(c1 * self.spacing)
                self.orientation[:3] = c2 / (self.spacing[0] * self.dimensions[2])
                self.orientation[3:] = c1 / (self.spacing[1] * self.dimensions[1])

        else:
            self.origin = origin

    def _find_skipped_slices(self, slice_direction):
        """
        Detects missing slices by checking irregular gaps in slice positions.
        """
        base_spacing = None
        for ii in range(len(self.image_set) - 1):
            position_1 = np.dot(slice_direction, self.image_set[ii].ImagePositionPatient)
            position_2 = np.dot(slice_direction, self.image_set[ii + 1].ImagePositionPatient)
            if ii == 0:
                base_spacing = position_2 - position_1
            if ii > 0 and np.abs(base_spacing - (position_2 - position_1)) > 0.01:
                self.unverified = 'Skipped'
                self.skipped_slice = ii + 1


class ReadXRay(object):
    """
    Reads and constructs 2D X-ray images (modalities DX and MG) from DICOM files. Tomosynthesis mammograms (3D MG) are
    not handled by this class.
    """

    def __init__(self, image_set, only_tags):
        """
        image_set: single DICOM image or list of DICOM datasets.
        only_tags: if True, loads only metadata without pixel data.


        """
        if isinstance(image_set, list):
            self.image_set = image_set
        else:
            self.image_set = [image_set]
        self.only_tags = only_tags

        self.unverified = 'Modality'
        self.skipped_slice = None
        self.sections = None
        self.rgb = False
        self.orientation = [1, 0, 0, 0, 1, 0]
        self.origin = np.asarray([0, 0, 0])
        self.image_matrix = np.identity(3, dtype=np.float32)

        self.modality = self.image_set[0].Modality

        self.filepaths = self.image_set[0].filename
        self.sops = self.image_set[0].SOPInstanceUID

        self.plane = self._compute_plane()
        self.dimensions = self._compute_dimensions()
        self.spacing = self._compute_spacing()

        self.array = None
        if not self.only_tags:
            self._compute_array()

        self.image_name = create_image_name(self.modality)

        image = Image(self)
        Data.image[self.image_name] = image
        Data.image_list += [self.image_name]

    def _compute_plane(self):
        """
        Determines the anatomical plane (Axial, Coronal, Sagittal) from the PatientOrientation DICOM tag. Defaults to
        Axial if not specified.
        """
        if 'PatientOrientation' in self.image_set[0]:
            orient = self.image_set[0].PatientOrientation
            if 'L' in orient or 'R' in orient:
                return 'Coronal'

            elif 'A' in orient or 'P' in orient:
                return 'Sagittal'

            else:
                return 'Axial'

        else:
            return 'Axial'

    def _compute_dimensions(self):
        """
        Defines image dimensions as [depth, height, width]. For 2D X-rays, one axis is set to 1 depending on the
        detected plane.
        """
        if self.plane == 'Axial':
            return np.asarray([1, self.image_set[0]['Rows'].value, self.image_set[0]['Columns'].value])

        elif self.plane == 'Coronal':
            return np.asarray([self.image_set[0]['Rows'].value, 1, self.image_set[0]['Columns'].value])

        else:
            return np.asarray([self.image_set[0]['Rows'].value, self.image_set[0]['Columns'].value, 1])

    def _compute_spacing(self):
        """
        Calculates voxel spacing [x, y, z]. Uses available DICOM tags (e.g. PixelSpacing, ImagerPixelSpacing).
        Sets slice thickness to 1 mm since X-rays are 2D. Adjusts spacing order based on the anatomical plane.
        """
        inplane_spacing = [1, 1]
        slice_thickness = 1

        if 'PixelSpacing' in self.image_set[0]:
            inplane_spacing = self.image_set[0].PixelSpacing

        elif 'ImagerPixelSpacing' in self.image_set[0]:
            inplane_spacing = self.image_set[0].ImagerPixelSpacing

        elif 'ContributingSourcesSequence' in self.image_set[0]:
            sequence = 'ContributingSourcesSequence'
            if 'DetectorElementSpacing' in self.image_set[0][sequence][0]:
                inplane_spacing = self.image_set[0][sequence][0]['DetectorElementSpacing']

        elif 'PerFrameFunctionalGroupsSequence' in self.image_set[0]:
            sequence = 'PerFrameFunctionalGroupsSequence'
            if 'PixelMeasuresSequence' in self.image_set[0][sequence][0]:
                inplane_spacing = self.image_set[0][sequence][0]['PixelMeasuresSequence'][0]['PixelSpacing']

        if self.plane == 'Axial':
            return np.asarray([inplane_spacing[1], inplane_spacing[0], slice_thickness])

        elif self.plane == 'Coronal':
            return np.asarray([inplane_spacing[1], slice_thickness, inplane_spacing[0]])

        else:
            return np.asarray([slice_thickness, inplane_spacing[1], inplane_spacing[0]])

    def _compute_array(self):
        """
        Reads the image pixel array (pixel_array) and converts it to int16. Deletes raw pixel data from memory after
        reading. If the PresentationLUTShape tag is 'Inverse', inverts grayscale values (16383 - array). Reshapes and
        flips the array according to the detected plane for consistent orientation.
        """
        self.array = self.image_set[0].pixel_array.astype('int16')
        del self.image_set[0].PixelData

        if 'PresentationLUTShape' in self.image_set[0] and self.image_set[0]['PresentationLUTShape'] == 'Inverse':
            self.array = 16383 - self.array

        if self.plane == 'Axial':
            self.array = self.array.reshape((1, self.array.shape[0], self.array.shape[1]))
        elif self.plane == 'Coronal':
            self.array = np.flip(np.flip(self.array.reshape((self.array.shape[0], 1, self.array.shape[1])), axis=0),
                                 axis=1)
        else:
            self.array = np.flip(self.array.reshape((self.array.shape[0], self.array.shape[1], 1)), axis=0)


class ReadRF(object):
    """
    Handles Radio Fluoroscopy (RF) image data from DICOM files. Builds a standardized representation containing
    metadata, orientation, and pixel data for integration into the global Data structure.
    """
    def __init__(self, image_set, only_tags):
        """
        image_set: a DICOM object or list of DICOM datasets.
        only_tags: if True, loads metadata only (no pixel data).
        """
        if isinstance(image_set, list):
            self.image_set = image_set
        else:
            self.image_set = [image_set]
        self.only_tags = only_tags

        self.unverified = 'Modality'
        self.skipped_slice = None
        self.sections = None
        self.rgb = False
        self.dimensions = None

        self.modality = self.image_set[0].Modality

        self.filepaths = self.image_set[0].filename
        self.sops = self.image_set[0].SOPInstanceUID
        self.orientation = [1, 0, 0, 0, 1, 0]
        self.origin = np.asarray([0, 0, 0])
        self.image_matrix = np.identity(3, dtype=np.float32)

        self.plane = self._compute_plane()

        self.array = None
        if not self.only_tags:
            self._compute_array()
        self.spacing = self._compute_spacing()

        self.image_name = create_image_name(self.modality)

        image = Image(self)
        Data.image[self.image_name] = image
        Data.image_list += [self.image_name]

    def _compute_plane(self):
        """
        Determines the anatomical viewing plane (Axial, Coronal, Sagittal) based on the DICOM PatientOrientation tag.
        Defaults to Axial if the tag is missing.
        """
        if 'PatientOrientation' in self.image_set[0]:
            orient = self.image_set[0].PatientOrientation
            if 'L' in orient or 'R' in orient:
                return 'Coronal'

            elif 'A' in orient or 'P' in orient:
                return 'Sagittal'

            else:
                return 'Axial'

        else:
            return 'Axial'

    def _compute_array(self):
        """
        Reads image pixel data into a NumPy array (int16). Deletes raw pixel data from memory to save space.
        Reshapes the array into a consistent 3D form, even for single 2D frames, depending on the detected plane.
        Saves the final shape as self.dimensions.
        """
        self.array = self.image_set[0].pixel_array.astype('int16')
        del self.image_set[0].PixelData

        if len(self.array.shape) < 3:
            if self.plane == 'Axial':
                self.array = self.array.reshape((self.array.shape[2], self.array.shape[0], self.array.shape[1]))
            elif self.plane == 'Coronal':
                self.array = self.array.reshape((self.array.shape[0], self.array.shape[2], self.array.shape[1]))
            else:
                self.array = self.array.reshape((self.array.shape[0], self.array.shape[1], self.array.shape[2]))

        self.dimensions = self.array.shape

    def _compute_spacing(self):
        """
        Determines voxel spacing using standard DICOM tags (PixelSpacing, ImagerPixelSpacing, etc.). If no tags are
        found, defaults to [1, 1, 1] (1 mm in each direction). Adjusts order of spacing values depending on the
        anatomical plane.
        """
        inplane_spacing = [1, 1]
        slice_thickness = 1

        if 'PixelSpacing' in self.image_set[0]:
            inplane_spacing = self.image_set[0].PixelSpacing

        elif 'ImagerPixelSpacing' in self.image_set[0]:
            inplane_spacing = self.image_set[0].ImagerPixelSpacing

        elif 'ContributingSourcesSequence' in self.image_set[0]:
            sequence = 'ContributingSourcesSequence'
            if 'DetectorElementSpacing' in self.image_set[0][sequence][0]:
                inplane_spacing = self.image_set[0][sequence][0]['DetectorElementSpacing']

        elif 'PerFrameFunctionalGroupsSequence' in self.image_set[0]:
            sequence = 'PerFrameFunctionalGroupsSequence'
            if 'PixelMeasuresSequence' in self.image_set[0][sequence][0]:
                inplane_spacing = self.image_set[0][sequence][0]['PixelMeasuresSequence'][0]['PixelSpacing']

        if self.plane == 'Axial':
            return np.asarray([inplane_spacing[1], inplane_spacing[0], slice_thickness])

        elif self.plane == 'Coronal':
            return np.asarray([inplane_spacing[1], slice_thickness, inplane_spacing[0]])

        else:
            return np.asarray([slice_thickness, inplane_spacing[1], inplane_spacing[0]])


class ReadUS(object):
    """
    Handles Ultrasound (US) images from DICOM files. Similar in structure to ReadXRay, but US images can include stacks
    of frames that do not represent true anatomical slices.
    """

    def __init__(self, image_set, only_tags):
        """
        image_set: a single DICOM US image or a list of images.
        only_tags: if True, loads only metadata without pixel data.
        """
        if isinstance(image_set, list):
            self.image_set = image_set
        else:
            self.image_set = [image_set]
        self.only_tags = only_tags

        self.unverified = 'Modality'
        self.base_position = None
        self.skipped_slice = None
        self.sections = None
        self.rgb = False

        self.modality = self.image_set[0].Modality

        self.filepaths = self.image_set[0].filename
        self.sops = self.image_set[0].SOPInstanceUID
        self.plane = 'Axial'
        self.orientation = [1, 0, 0, 0, 1, 0]
        self.origin = np.asarray([0, 0, 0])
        self.image_matrix = np.identity(3, dtype=np.float32)
        self.dimensions = np.asarray([1, self.image_set[0]['Rows'].value, self.image_set[0]['Columns'].value])

        self.array = None
        if not self.only_tags:
            self._compute_array()
        self.spacing = self._compute_spacing()

        self.image_name = create_image_name(self.modality)

        image = Image(self)
        Data.image[self.image_name] = image
        Data.image_list += [self.image_name]

    def _compute_array(self):
        """
        Reads pixel data and converts to NumPy array (uint8). Reshapes 2D images to 3D arrays with a single frame.
        For multi-frame images:
            Checks for uniform frames and extracts the first channel.
            Updates self.dimensions based on the number of frames.
        """
        us_data = np.asarray(self.image_set[0].pixel_array)
        del self.image_set[0].PixelData

        if len(us_data.shape) == 2:
            us_data = us_data.reshape((1, us_data.shape[0], us_data.shape[1]))

        if len(us_data.shape) == 3:
            us_binary = (1 * (np.std(us_data, axis=2) == 0) == 1)
            self.array = (us_binary * us_data[:, :, 0]).astype('uint8')
            if len(self.array.shape) == 2:
                self.array = np.expand_dims(self.array, axis=0)

        else:
            us_binary = (1 * (np.std(us_data, axis=3) == 0) == 1)
            self.array = (us_binary * us_data[:, :, :, 0]).astype('uint8')

        if len(self.array.shape) == 3:
            self.dimensions[0] = self.array.shape[0]

    def _compute_spacing(self):
        """
        Determines pixel spacing using standard DICOM tags (PixelSpacing, DetectorElementSpacing, etc.) or
        ultrasound-specific sequences. Defaults to [1, 1, 1] if tags are missing. Returns spacing in [x, y, z] order
        with a default slice thickness of 1 mm.
        """
        inplane_spacing = [1, 1]
        slice_thickness = 1

        if 'PixelSpacing' in self.image_set[0]:
            inplane_spacing = self.image_set[0].PixelSpacing

        elif 'ContributingSourcesSequence' in self.image_set[0]:
            sequence = 'ContributingSourcesSequence'
            if 'DetectorElementSpacing' in self.image_set[0][sequence][0]:
                inplane_spacing = self.image_set[0][sequence][0]['DetectorElementSpacing']

        elif 'PerFrameFunctionalGroupsSequence' in self.image_set[0]:
            sequence = 'PerFrameFunctionalGroupsSequence'
            if 'PixelMeasuresSequence' in self.image_set[0][sequence][0]:
                inplane_spacing = self.image_set[0][sequence][0]['PixelMeasuresSequence'][0]['PixelSpacing']

        elif 'SequenceOfUltrasoundRegions' in self.image_set[0]:
            if 'PhysicalDeltaX' in self.image_set[0].SequenceOfUltrasoundRegions[0]:
                inplane_spacing = [10 * np.round(self.image_set[0].SequenceOfUltrasoundRegions[0].PhysicalDeltaY, 4),
                                   10 * np.round(self.image_set[0].SequenceOfUltrasoundRegions[0].PhysicalDeltaX, 4)]

        return np.asarray([inplane_spacing[1], inplane_spacing[0], slice_thickness])


class ReadRTStruct(object):
    """
    Handles Radiotherapy Structure Set (RTSTRUCT) DICOM files. Extracts ROI (Region of Interest) contours and POI
    (Points of Interest), matches them to a corresponding image series, and prepares them for integration with the
    global Data structure.
    """
    def __init__(self, image_set, only_tags):
        """
        image_set: a DICOM RTSTRUCT object.
        only_tags: if True, only metadata is loaded (no contour/point coordinates).
        """
        self.image_set = image_set
        self.only_tags = only_tags

        self.series_uid = self._get_series_uid()
        self.filepaths = self.image_set.filename

        self._properties = self._get_properties()
        self.roi_names = [prop[1] for prop in self._properties if prop[3].lower() == 'closed_planar']
        self.roi_colors = [prop[2] for prop in self._properties if prop[3].lower() == 'closed_planar']
        self.poi_names = [prop[1] for prop in self._properties if prop[3].lower() == 'point']
        self.poi_colors = [prop[2] for prop in self._properties if prop[3].lower() == 'point']

        if len(self.roi_names) > 0 or len(self.poi_names) > 0:
            self.match_image_name = self._match_with_image()

            self.contours = []
            self.points = []
            if not self.only_tags:
                self._structure_positions()
        else:
            self.match_image_name = None

    def _get_series_uid(self):
        """
        Reads the SeriesInstanceUID of the image series referenced by the RTSTRUCT.
        """
        study = 'RTReferencedStudySequence'
        series = 'RTReferencedSeriesSequence'
        ref = self.image_set.ReferencedFrameOfReferenceSequence

        return ref[0][study][0][series][0]['SeriesInstanceUID'].value

    def _get_properties(self):
        """
        Iterates through ROIContourSequence. Collects ROI/POI names, colors, geometric type, and linked SOPInstanceUIDs.
        Assigns random colors if none are provided.
        """
        tracker = []
        sop = []
        names = []
        colors = []
        geometric = []
        for ii, s in enumerate(self.image_set.ROIContourSequence):
            if hasattr(self.image_set.StructureSetROISequence[ii], 'ROIName'):
                if hasattr(s, 'ContourSequence'):
                    tracker += [ii]
                    names += [self.image_set.StructureSetROISequence[ii].ROIName]
                    geometric += [s['ContourSequence'][0]['ContourGeometricType'].value]

                    slice_sop = []
                    if geometric[-1].lower() == 'closed_planar':
                        for seq in s['ContourSequence']:
                            slice_sop += [seq['ContourImageSequence'][0]['ReferencedSOPInstanceUID'].value]
                    else:
                        if 'ContourImageSequence' in s['ContourSequence'][0]:
                            slice_sop = [s['ContourSequence'][0]['ContourImageSequence'][0]['ReferencedSOPInstanceUID'].value]
                        else:
                            slice_sop = []
                    sop += [slice_sop]

                    if hasattr(s, 'ROIDisplayColor'):
                        colors += [s.ROIDisplayColor]
                    else:
                        colors += [[np.random.randint(0, 256), np.random.randint(0, 256), np.random.randint(0, 256)]]

        properties = []
        for ii in range(len(names)):
            properties += [[tracker[ii], names[ii], colors[ii], geometric[ii], sop[ii]]]

        return properties

    def _match_with_image(self):
        """
        Searches Data.image for an image with a matching series UID and SOPInstanceUID. Returns the matched image name
        or None if no match exists.
        """
        match_image_name = None

        for image_name in Data.image:
            if self.series_uid == Data.image[image_name].series_uid:
                if self._properties[0][4][0] in Data.image[image_name].sops:
                    match_image_name = image_name

        return match_image_name

    def _structure_positions(self):
        """
        Extracts the 3D coordinates of contours and points. Stores planar ROIs in self.contours and points in
        self.points.
        """
        sequences = self.image_set.ROIContourSequence
        for prop in self._properties:
            seq = sequences[prop[0]]

            contour_list = []
            for c in seq.ContourSequence:
                contour_hold = np.round(np.array(c['ContourData'].value), 3)
                contour = contour_hold.reshape(int(len(contour_hold) / 3), 3)
                contour_list.append(contour)

            if prop[3].lower() == 'closed_planar':
                self.contours += [contour_list]
            else:
                self.points += contour_list


class ReadREG(object):
    def __init__(self, image_set, only_tags):
        if isinstance(image_set, list):
            self.image_set = image_set
        else:
            self.image_set = [image_set]
        self.only_tags = only_tags

        self.reference_name = None
        self.reference_series = self.image_set[0].ReferencedSeriesSequence[0].SeriesInstanceUID
        self.reference_sops = [sop.ReferencedSOPInstanceUID for sop in
                               self.image_set[0].ReferencedSeriesSequence[0].ReferencedInstanceSequence]

        self.moving_name = None
        if len(self.image_set[0].ReferencedSeriesSequence) == 2:
            self.moving_series = self.image_set[0].ReferencedSeriesSequence[1].SeriesInstanceUID
            self.moving_sops = [sop.ReferencedSOPInstanceUID for sop in
                                self.image_set[0].ReferencedSeriesSequence[1].ReferencedInstanceSequence]
        else:
            sequence = self.image_set[0].StudiesContainingOtherReferencedInstancesSequence[0].ReferencedSeriesSequence[0]
            self.moving_series = sequence.SeriesInstanceUID
            self.moving_sops = [sop.ReferencedSOPInstanceUID for sop in sequence.ReferencedInstanceSequence]

        self.spacing = None
        self.dimensions = None
        self.origin = None

        self.reference_matrix = None
        self.moving_matrix = None
        self.dvf_matrix = None
        self.dvf = None

        self.registration_name = None
        if 'DeformableRegistrationSequence' in self.image_set[0]:
            self._compute_rigid(deformable=True)
            self._compute_dvf()
            self._create_name(deformable=True)
            self._create_registration(deformable=True)
        else:
            self._compute_rigid()
            self._create_name()
            self._create_registration()

    def _compute_sops(self):
        pass

    def _compute_rigid(self, deformable=False):
        if deformable:
            matrix = self.image_set[0].DeformableRegistrationSequence[0].PreDeformationMatrixRegistrationSequence[0][
                0x3006, 0x00C6].value
            orientation = self.image_set[0].DeformableRegistrationSequence[0].DeformableRegistrationGridSequence[0].ImageOrientationPatient
            row_direction = orientation[:3]
            column_direction = orientation[3:]
            slice_direction = np.cross(row_direction, column_direction)

            self.dvf_matrix = np.identity(3, dtype=np.float32)
            self.dvf_matrix[0, :3] = row_direction
            self.dvf_matrix[1, :3] = column_direction
            self.dvf_matrix[2, :3] = slice_direction
            self.moving_matrix = np.linalg.inv(np.asarray(matrix).reshape(4, 4))
        else:
            self.reference_matrix = self.image_set[0].RegistrationSequence[1]['MatrixRegistrationSequence'][0]['MatrixSequence'][0][0x3006, 0x00C6].value
            matrix = self.image_set[0].RegistrationSequence[1]['MatrixRegistrationSequence'][0]['MatrixSequence'][0][0x3006, 0x00C6].value
            self.moving_matrix = np.linalg.inv(np.asarray(matrix).reshape(4, 4))

    def _compute_dvf(self):
        self.origin = self.image_set[0].DeformableRegistrationSequence[0].DeformableRegistrationGridSequence[
            0].ImagePositionPatient
        self.dimensions = np.flip(
            self.image_set[0].DeformableRegistrationSequence[0].DeformableRegistrationGridSequence[0].GridDimensions)
        self.spacing = self.image_set[0].DeformableRegistrationSequence[0].DeformableRegistrationGridSequence[0].GridResolution

        dvf_raw = self.image_set[0].DeformableRegistrationSequence[0].DeformableRegistrationGridSequence[
            0].VectorGridData
        dvf_unstructured = unpack(f"<{len(dvf_raw) // 4}f", dvf_raw)
        self.dvf = np.reshape(dvf_unstructured, list(self.dimensions) + [3, ])

        del self.image_set[0].DeformableRegistrationSequence[0].DeformableRegistrationGridSequence[0].VectorGridData

    def _create_name(self, deformable=False):
        for image_name in Data.image_list:
            if self.reference_sops[0] in Data.image[image_name].sops:
                self.reference_name = image_name

            elif self.moving_sops[0] in Data.image[image_name].sops:
                self.moving_name = image_name

        if deformable:
            registration = 'DVF_'
        else:
            registration = ''

        if self.reference_name is None and self.moving_name is None:
            registration_name = registration + '_Unknown'
        else:
            registration_name = registration + self.reference_name + '_' + self.moving_name

        if deformable:
            if registration_name in Data.deformable_list:
                n = 0
                while n > -1:
                    n += 1
                    new_name = copy.deepcopy(registration_name + '_' + str(n))
                    if new_name not in Data.deformable_list:
                        self.registration_name = new_name
                        n = -100
            else:
                self.registration_name = registration_name
        else:
            if registration_name in Data.rigid_list:
                n = 0
                while n > -1:
                    n += 1
                    new_name = copy.deepcopy(registration_name + '_' + str(n))
                    if new_name not in Data.rigid_list:
                        self.registration_name = new_name
                        n = -100
            else:
                self.registration_name = registration_name

    def _create_registration(self, deformable=False):
        if deformable:
            Deformable(self.dvf, self.origin, self.spacing, self.dimensions, rigid_matrix=self.moving_matrix,
                       dvf_matrix=self.dvf_matrix, registration_name=self.registration_name,
                       reference_name=self.reference_name, moving_name=self.moving_name,
                       reference_sops=self.reference_sops, moving_sops=self.moving_sops)

        elif self.reference_name is not None and self.moving_name is not None:
            Rigid(self.reference_name, self.moving_name, rigid_name=self.registration_name,
                  reference_sops=self.reference_sops, moving_sops=self.moving_sops,
                  reference_matrix=self.reference_matrix, matrix=self.moving_matrix)


class ReadRTDose(object):
    """
    Handles Radiotherapy Dose (RTDOSE) DICOM files. Reads dose arrays, determines spacing, orientation, and dimensions,
    and integrates the dose into the global Data structure.
    """
    def __init__(self, image_set, only_tags):
        """
        image_set: one or more DICOM RTDOSE objects.
        only_tags: if True, only metadata is loaded (no dose array computation).
        """
        if isinstance(image_set, list):
            self.image_set = image_set
        else:
            self.image_set = [image_set]
        self.only_tags = only_tags

        self.unverified = None
        self.base_position = None
        self.skipped_slice = None
        self.sections = None
        self.modality = 'RTDOSE'

        self.filepaths = [image.filename for image in self.image_set]
        self.sops = [image.SOPInstanceUID for image in self.image_set]

        self.array = None
        if not self.only_tags:
            self._compute_array()

        self.orientation = self._compute_orientation()
        self.plane = self._compute_plane()
        self.spacing = self._compute_spacing()
        self.dimensions = self._compute_dimensions()
        self._verify_axial_orientation()

        self.image_matrix = self._compute_image_matrix()
        self.dose_name = create_dose_name(self.modality)

        dose = Dose(self)
        Data.dose[self.dose_name] = dose
        Data.dose_list += [self.dose_name]

    def _compute_array(self):
        """
        Combines slices into a 3D dose array. Applies DoseGridScaling if present. Deletes original pixel data.
        """

        if (0x3004, 0x000E) in self.image_set[0]:
            slope = self.image_set[0].DoseGridScaling
        else:
            slope = 1

        dose_array = (self.image_set[0].pixel_array * slope)

        self.array = np.asarray(dose_array)
        if len(dose_array.shape) == 2:
            self.array.reshape((1, self.array.shape[0], self.array.shape[1]))

        del self.image_set[0].PixelData

    def _compute_orientation(self):
        """
        Reads orientation from ImageOrientationPatient or functional group sequences. Flags unverified if orientation
        cannot be determined.
        """
        orientation = np.asarray([1, 0, 0, 0, 1, 0])
        if 'ImageOrientationPatient' in self.image_set[0]:
            orientation = np.asarray(self.image_set[0]['ImageOrientationPatient'].value)

        else:
            if 'SharedFunctionalGroupsSequence' in self.image_set[0]:
                seq_str = 'SharedFunctionalGroupsSequence'
                if 'PlaneOrientationSequence' in self.image_set[0][0][seq_str][0]:
                    plane_str = 'PlaneOrientationSequence'
                    image_str = 'ImageOrientationPatient'
                    orientation = np.asarray(self.image_set[0][0][seq_str][0][plane_str][0][image_str].value)

                else:
                    self.unverified = 'Orientation'

            else:
                self.unverified = 'Orientation'

        return orientation

    def _compute_plane(self):
        """
        Determines the anatomical plane (Axial, Coronal, Sagittal) based on orientation vectors.
        """
        x = np.abs(self.orientation[0]) + np.abs(self.orientation[3])
        y = np.abs(self.orientation[1]) + np.abs(self.orientation[4])
        z = np.abs(self.orientation[2]) + np.abs(self.orientation[5])

        if x < y and x < z:
            return 'Sagittal'
        elif y < x and y < z:
            return 'Coronal'
        else:
            return 'Axial'

    def _compute_spacing(self):
        """
        Computes 3-axis voxel spacing from pixel spacing and slice thickness. Detects skipped slices and adjusts spacing
        if slices are irregularly spaced (_find_skipped_slices).
        """
        inplane_spacing = [1, 1]
        slice_thickness = np.double(self.image_set[0].SliceThickness)

        if 'PixelSpacing' in self.image_set[0]:
            inplane_spacing = self.image_set[0].PixelSpacing

        elif 'ContributingSourcesSequence' in self.image_set[0]:
            sequence = 'ContributingSourcesSequence'
            if 'DetectorElementSpacing' in self.image_set[0][sequence][0]:
                inplane_spacing = self.image_set[0][sequence][0]['DetectorElementSpacing']

        elif 'PerFrameFunctionalGroupsSequence' in self.image_set[0]:
            sequence = 'PerFrameFunctionalGroupsSequence'
            if 'PixelMeasuresSequence' in self.image_set[0][sequence][0]:
                inplane_spacing = self.image_set[0][sequence][0]['PixelMeasuresSequence'][0]['PixelSpacing']

        if len(self.image_set) > 1:
            row_direction = self.orientation[:3]
            column_direction = self.orientation[3:]
            slice_direction = np.cross(row_direction, column_direction)

            first = np.dot(slice_direction, self.image_set[0].ImagePositionPatient)
            second = np.dot(slice_direction, self.image_set[1].ImagePositionPatient)
            last = np.dot(slice_direction, self.image_set[-1].ImagePositionPatient)
            first_last_spacing = np.asarray((last - first) / (len(self.image_set) - 1))
            if np.abs((second - first) - first_last_spacing) > 0.01:
                if not self.only_tags:
                    self._find_skipped_slices(slice_direction)
                slice_thickness = second - first
            else:
                slice_thickness = np.asarray((last - first) / (len(self.image_set) - 1))

        if self.plane == 'Axial':
            return np.asarray([inplane_spacing[1], inplane_spacing[0], slice_thickness])

        elif self.plane == 'Coronal':
            return np.asarray([inplane_spacing[1], slice_thickness, inplane_spacing[0]])

        else:
            return np.asarray([slice_thickness, inplane_spacing[1], inplane_spacing[0]])

    def _compute_dimensions(self):
        """
        Sets dimensions based on array shape and plane orientation.
        """
        shape = self.array.shape
        if self.plane == 'Axial':
            return np.asarray([shape[0], shape[1], shape[2]])

        elif self.plane == 'Coronal':
            return np.asarray([shape[1], shape[0], shape[2]])

        else:
            return np.asarray([shape[1], shape[2], shape[0]])

    def _compute_image_matrix(self):
        """
        Constructs a 3×3 image matrix using row, column, and slice directions.
        """
        row_direction = self.orientation[:3]
        column_direction = self.orientation[3:]
        slice_direction = np.cross(row_direction, column_direction)

        mat = np.identity(3, dtype=np.float32)
        mat[0, :3] = row_direction
        mat[1, :3] = column_direction
        mat[2, :3] = slice_direction

        return mat

    def _verify_axial_orientation(self):
        """
        Checks and corrects the origin and orientation of the dose array to ensure proper alignment. Rotates or flips
        the array as needed.
        """
        shape = self.array.shape
        if self.plane == 'Axial':
            spacing = self.spacing
        elif self.plane == 'Coronal':
            spacing = [self.spacing[0], self.spacing[2], self.spacing[1]]
        else:
            spacing = [self.spacing[1], self.spacing[2], self.spacing[0]]

        slices = shape[0] - 1
        y = shape[1] - 1
        x = shape[2] - 1

        origin = np.asarray(self.image_set[0]['ImagePositionPatient'].value)

        row_direction = self.orientation[:3]
        column_direction = self.orientation[3:]
        slice_direction = np.cross(row_direction, column_direction)

        corners = np.zeros((8, 3))
        corners[0] = origin
        corners[1] = origin + (x * spacing[0] * row_direction)
        corners[2] = origin + (y * spacing[1] * column_direction)
        corners[3] = (origin + (x * spacing[0] * row_direction) + (y * spacing[1] * column_direction))

        corners[4] = origin + (slices * spacing[2] * slice_direction)
        corners[5] = (origin + (slices * spacing[2] * slice_direction) + (x * spacing[0] * row_direction))
        corners[6] = (origin + (slices * spacing[2] * slice_direction) + (y * spacing[1] * column_direction))
        corners[7] = (origin + (slices * spacing[2] * slice_direction) + (x * spacing[0] * row_direction) +
                      (y * spacing[1] * column_direction))

        corner_idx = np.argmin(np.sum(corners, axis=1))
        if corner_idx != 0:
            self.origin = corners[corner_idx]
            if self.plane == "Axial":
                if corner_idx == 1:
                    self.array = np.rot90(self.array, 1, (1, 2))
                elif corner_idx == 2:
                    self.array = np.rot90(self.array, 3, (1, 2))
                else:
                    self.array = np.rot90(self.array, 2, (1, 2))

                if corner_idx < 4:
                    square = corners[:4, :]
                else:
                    square = corners[4:, :]

            elif self.plane == 'Coronal':
                self.array = np.rot90(self.array, 1, (0, 1))

                s1 = np.argsort(corners[:4, 2])
                s2 = np.argsort(corners[4:, 2]) + 4

                square = [corners[s1[0]], corners[s1[1]], corners[s2[0]], corners[s2[1]]]

            else:
                self.array = np.flip(np.rot90(self.array, 1, (0, 1)).transpose(0, 2, 1), axis=2)

                s1 = np.argsort(corners[:4, 2])
                s2 = np.argsort(corners[4:, 2]) + 4

                square = [corners[s1[0]], corners[s1[1]], corners[s2[0]], corners[s2[1]]]

            distances = np.asarray([np.linalg.norm(corners[corner_idx, :] - s) for s in square])
            sorted_args = np.argsort(distances)

            c1 = square[sorted_args[1]] - corners[corner_idx]
            c2 = square[sorted_args[2]] - corners[corner_idx]

            if np.abs(c1[0]) > np.abs(c2[0]):
                # self.orientation[:3] = c1 / self.spacing / np.flip(self.dimensions - 1)
                # self.orientation[3:] = c2 / self.spacing / np.flip(self.dimensions - 1)
                # self.orientation[:3] = c1 * self.spacing / np.linalg.norm(c1 * self.spacing)
                # self.orientation[3:] = c2 * self.spacing / np.linalg.norm(c2 * self.spacing)
                self.orientation[:3] = c1 / (self.spacing[0] * self.dimensions[2])
                self.orientation[3:] = c2 / (self.spacing[1] * self.dimensions[1])
            else:
                # self.orientation[:3] = c2 / self.spacing / np.flip(self.dimensions - 1)
                # self.orientation[3:] = c1 / self.spacing / np.flip(self.dimensions - 1)
                # self.orientation[:3] = c2 * self.spacing / np.linalg.norm(c2 * self.spacing)
                # self.orientation[3:] = c1 * self.spacing / np.linalg.norm(c1 * self.spacing)
                self.orientation[:3] = c2 / (self.spacing[0] * self.dimensions[2])
                self.orientation[3:] = c1 / (self.spacing[1] * self.dimensions[1])

        else:
            self.origin = origin

    def _find_skipped_slices(self, slice_direction):
        """
        Detects skipped or irregular slices along the slice direction and flags them.
        """
        base_spacing = None
        for ii in range(len(self.image_set) - 1):
            position_1 = np.dot(slice_direction, self.image_set[ii].ImagePositionPatient)
            position_2 = np.dot(slice_direction, self.image_set[ii + 1].ImagePositionPatient)
            if ii == 0:
                base_spacing = position_2 - position_1
            if ii > 0 and np.abs(base_spacing - (position_2 - position_1)) > 0.01:
                self.unverified = 'Skipped'
                self.skipped_slice = ii + 1


def create_image_name(modality):
    """
    Generate a unique, sequential name for an image based on its modality.
    """
    idx = len(Data.image_list)
    if idx < 9:
        image_name = modality + ' 0' + str(1 + idx)
    else:
        image_name = modality + ' ' + str(1 + idx)

    return image_name


def create_dose_name(modality):
    """
    Generate a unique, sequential name for a dose based on its modality.
    """
    idx = len(Data.dose_list)
    if idx < 9:
        image_name = modality + ' 0' + str(1 + idx)
    else:
        image_name = modality + ' ' + str(1 + idx)

    return image_name



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:
    Under construction

Structure:

"""

import pytetwild

import numpy as np
import pyvista as pv


class Volume(object):
    """
    Converts a surface mesh into a volumetric tetrahedral mesh using PyTetWild and PyVista.
    Provides methods to create the volumetric mesh and save it to a file.
    """
    def __init__(self, surface_mesh):
        self.surface_mesh = surface_mesh
        self.mesh = None

    def create(self, edge_length=.02):
        """
        Generates a tetrahedral volumetric mesh from the surface mesh.

        Parameters
        ----------
        edge_length : float
            Target edge length factor for tetrahedralization. Smaller values yield finer meshes.
        """
        base_mesh = pytetwild.tetrahedralize_pv(self.surface_mesh, edge_length_fac=edge_length, optimize=True)
        tetra_connect = base_mesh.cell_connectivity.reshape(int(len(base_mesh.cell_connectivity) / 4), 4)
        tetra_reshape = np.asarray([4 * np.ones((tetra_connect.shape[0])),
                                    tetra_connect[:, 3],
                                    tetra_connect[:, 1],
                                    tetra_connect[:, 2],
                                    tetra_connect[:, 0]], dtype=pv.ID_TYPE).T
        tetra_reformat = tetra_reshape.reshape(int(len(base_mesh.cell_connectivity) * 1.25))
        tetra_total = int(len(base_mesh.cell_connectivity) / 4)
        cell_types = np.array([pv.CellType.TETRA] * tetra_total)
        self.mesh = pv.UnstructuredGrid(tetra_reformat, cell_types, np.asarray(base_mesh.points))

    def write(self, path):
        """
        Saves the tetrahedral mesh to a file.

        Parameters
        ----------
        path : str
            Path to save the mesh file. Can be .vtu or other PyVista-supported formats.
        """
        self.mesh.save(path, binary=False)



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:

Structure:

"""

import vtk
import numpy as np
import SimpleITK as sitk

from scipy.spatial.transform import Rotation

from ..utils.mesh.surface import Refinement
from ..utils.conversion import ContourToDiscreteMesh


class Roi(object):
    def __init__(self, image, position=None, name=None, color=None, visible=False, filepaths=None, plane=None):
        self.image = image

        self.name = name
        self.visible = visible
        self.color = color
        self.filepaths = filepaths

        if plane is not None:
            self.plane = plane
        else:
            self.plane = self.image.plane

        if position is not None:
            self.contour_position = position
            self.contour_pixel = self.convert_position_to_pixel(position)
        else:
            self.contour_position = None
            self.contour_pixel = None

        self.mesh = None
        self.volume = None
        self.com = None
        self.bounds = None

        self.fixed_name = False
        self.visual = {'2d': None, '3d': None, 'opacity': None, 'multicolor': None}
        self.misc = {}

    def add_mesh(self, mesh):
        self.mesh = mesh

        self.volume = mesh.volume
        self.com = mesh.center
        self.bounds = mesh.bounds

    def clear(self):
        self.contour_position = None
        self.contour_pixel = None

        self.mesh = None
        self.volume = None
        self.com = None
        self.bounds = None

        self.fixed_name = False
        self.visual = {'2d': None, '3d': None, 'opacity': None, 'multicolor': None}
        self.misc = {}

    def convert_position_to_pixel(self, position=None):
        position_to_pixel_matrix = self.image.display.compute_matrix_position_to_pixel()

        pixel = []
        for ii, pos in enumerate(position):
            p_concat = np.concatenate((pos, np.ones((pos.shape[0], 1))), axis=1)
            pixel_3_axis = p_concat.dot(position_to_pixel_matrix.T)[:, :3]
            pixel += [np.vstack((pixel_3_axis, pixel_3_axis[0, :]))]

        return pixel

    def convert_pixel_to_position(self, pixel=None):
        pixel_to_position_matrix = self.image.display.compute_matrix_pixel_to_position()

        position = []
        for ii, pix in enumerate(pixel):
            p_concat = np.concatenate((pix, np.ones((pix.shape[0], 1))), axis=1)
            position += [p_concat.dot(pixel_to_position_matrix.T)[:, :3]]

        return position

    def create_mesh(self, smoothing_iterations=20, smoothing_relaxation=.5, smoothing_distance=1):
        meshing = ContourToDiscreteMesh(contour_pixel=self.contour_pixel,
                                        spacing=self.image.spacing,
                                        origin=self.image.origin,
                                        dimensions=self.image.dimensions,
                                        matrix=self.image.matrix,
                                        plane=self.plane)
        self.mesh = meshing.compute_mesh(smoothing_iterations=smoothing_iterations,
                                         smoothing_relaxation=smoothing_relaxation,
                                         smoothing_distance=smoothing_distance)
        self.volume = self.mesh.volume
        self.com = self.mesh.center
        self.bounds = self.mesh.bounds

    def create_discrete_mesh(self):
        meshing = ContourToDiscreteMesh(contour_pixel=self.contour_pixel,
                                        spacing=self.image.spacing,
                                        origin=self.image.origin,
                                        dimensions=self.image.dimensions,
                                        matrix=self.image.matrix,
                                        plane=self.plane)
        self.mesh = meshing.compute_mesh(discrete=True)

        self.volume = self.mesh.volume
        self.com = self.mesh.center
        self.bounds = self.mesh.bounds

    def create_display_mesh(self, iterations=20, angle=60, passband=0.001):
        refine = Refinement(self.mesh)
        self.mesh = refine.smooth(iterations=iterations, angle=angle, passband=passband)

    def create_decimate_mesh(self, percent=None, set_mesh=False):
        if percent is None:
            points = np.round(10 * np.sqrt(self.mesh.number_of_points))
            percent = 1 - (points / self.mesh.number_of_points)

        mesh = self.mesh.decimate(percent)
        if set_mesh:
            self.mesh = mesh

        return mesh

    def create_cluster_mesh(self, points=None, set_mesh=False):
        refine = Refinement(self.mesh)
        mesh = refine.cluster(points=points)
        if set_mesh:
            self.mesh = mesh

        return mesh

    def compute_contour(self, slice_location, offset=0):
        contour_list = []
        if self.contour_pixel is not None:
            if self.plane == 'Axial':
                roi_z = [np.round(c[0, 2]).astype(int) for c in self.contour_pixel]
                keep_idx = np.argwhere(np.asarray(roi_z) == slice_location)

                if len(keep_idx) > 0:
                    for ii, idx in enumerate(keep_idx):
                        contour_corrected = np.vstack((self.contour_pixel[idx[0]][:, 0:2] + offset,
                                                       self.contour_pixel[idx[0]][0, 0:2] + offset))
                        contour_list += [contour_corrected]

            elif self.plane == 'Coronal':
                roi_y = [np.round(c[0, 1]).astype(int) for c in self.contour_pixel]
                keep_idx = np.argwhere(np.asarray(roi_y) == slice_location)

                if len(keep_idx) > 0:
                    for ii, idx in enumerate(keep_idx):
                        pixel_reshape = np.column_stack((self.contour_pixel[idx[0]][:, 0] + offset,
                                                         self.contour_pixel[idx[0]][:, 2] + offset))
                        stack = np.asarray([self.contour_pixel[idx[0]][0, 0], self.contour_pixel[idx[0]][0, 2]])
                        contour_corrected = np.vstack((pixel_reshape, stack))
                        contour_list += [contour_corrected]

            else:
                roi_x = [np.round(c[0, 0]).astype(int) for c in self.contour_pixel]
                keep_idx = np.argwhere(np.asarray(roi_x) == slice_location)

                if len(keep_idx) > 0:
                    for ii, idx in enumerate(keep_idx):
                        contour_corrected = np.vstack((self.contour_pixel[idx[0]][:, 1:] + offset,
                                                       self.contour_pixel[idx[0]][0, 1:] + offset))
                        contour_list += [contour_corrected]

        return contour_list

    def compute_mask(self):
        mask = ContourToDiscreteMesh(contour_pixel=self.contour_pixel,
                                     spacing=self.image.spacing,
                                     origin=self.image.origin,
                                     dimensions=self.image.dimensions,
                                     matrix=self.image.matrix,
                                     plane=self.plane)

        return mask.mask

    def compute_mesh_slice(self, location=None, slice_plane=None, offset=0, return_pixel=False):
        matrix = np.linalg.inv(self.image.display.matrix)
        if slice_plane == 'Axial':
            normal = matrix[:3, 2]
        elif slice_plane == 'Coronal':
            normal = matrix[:3, 1]
        else:
            normal = matrix[:3, 0]

        roi_slice = self.mesh.slice(normal=normal, origin=location)

        if return_pixel:
            if roi_slice.number_of_points > 0:
                roi_strip = roi_slice.strip(max_length=10000000)

                # colors = None
                # if self.multi_color:
                #     strip_colors = roi_strip['colors']
                #
                #     position = []
                #     colors = []
                #     for cell in roi_strip.cell:
                #         position += [np.asarray(roi_strip.points[cell.point_ids])]
                #         colors += [np.asarray(strip_colors[cell.point_ids])]
                #
                # else:
                #     position = []
                #     for cell in roi_strip.cell:
                #         position += [np.asarray(roi_strip.points[cell.point_ids])]

                colors = None
                position = [np.asarray(c.points) for c in roi_strip.cell]
                pixels = self.convert_position_to_pixel(position=position)
                pixel_corrected = []
                for pixel in pixels:

                    if slice_plane in 'Axial':
                        pixel_reshape = pixel[:, :2] + offset
                        pixel_corrected += [np.asarray([pixel_reshape[:, 0], pixel_reshape[:, 1]]).T]

                    elif slice_plane == 'Coronal':
                        pixel_reshape = np.column_stack((pixel[:, 0] + offset, pixel[:, 2] + offset))
                        pixel_corrected += [pixel_reshape]

                    else:
                        pixel_reshape = pixel[:, 1:] + offset
                        pixel_corrected += [pixel_reshape]

                return pixel_corrected, colors

            else:
                return [], None

        else:
            colors = None
            # if self.multi_color:
            #     colors = roi_slice['colors']

            return roi_slice, colors

    def create_sitk_mask(self):
        mask = self.compute_mask()

        matrix_flat = self.image.matrix.flatten(order='F')
        sitk_mask = sitk.GetImageFromArray(mask.T)
        sitk_mask.SetDirection([float(mat) for mat in matrix_flat])
        sitk_mask.SetOrigin(self.image.origin)
        sitk_mask.SetSpacing(self.image.spacing)

        return sitk_mask

    def update_pixel(self, pixel, plane='Axial'):
        self.plane = plane
        self.contour_pixel = pixel
        self.contour_position = self.convert_pixel_to_position(pixel=pixel)

        self.create_discrete_mesh()
        self.create_display_mesh()

    def update_mesh(self, mesh):
        self.mesh = mesh
        self.volume = mesh.volume
        self.com = mesh.center
        self.bounds = mesh.bounds

        self.contour_pixel = None
        self.contour_position = None



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org
"""

import os
import psutil

from .read import DicomReader, MhdReader, StlReader, VtkReader, ThreeMfReader


def check_memory(files):
    """
    Estimate available memory after accounting for file sizes in different formats.

    Parameters
    ----------
    files : dict
        Dictionary of file lists grouped by type. Keys: 'Dicom', 'Nifti', 'Raw', 'Stl', 'Vtk', '3mf'

    Returns
    -------
    float
        Available memory in GB after subtracting the total file sizes.
    """

    dicom_size = 0
    for file in files['Dicom']:
        dicom_size = dicom_size + os.path.getsize(file)

    nifti_size = 0
    for file in files['Nifti']:
        nifti_size = nifti_size + os.path.getsize(file)

    raw_size = 0
    for file in files['Raw']:
        raw_size = raw_size + os.path.getsize(file)

    stl_size = 0
    for file in files['Stl']:
        stl_size = stl_size + os.path.getsize(file)

    vtk_size = 0
    for file in files['Vtk']:
        vtk_size = vtk_size + os.path.getsize(file)

    mf3_size = 0
    for file in files['3mf']:
        mf3_size = mf3_size + os.path.getsize(file)

    total_size = dicom_size + raw_size + nifti_size + stl_size + vtk_size + mf3_size
    available_memory = psutil.virtual_memory()[1]
    return (available_memory - total_size) / 1000000000


def file_parsar(folder_path=None, file_list=None, exclude_files=None):
    """
    Walk through folders or file list and sort files into categories based on extension.

    Supports:
        DICOM (.dcm)
        MHD (.mhd)
        RAW (.raw)
        Nifti (.nii.gz)
        STL (.stl)
        VTK (.vtk)
        3MF (.3mf)
        No extension

    Parameters
    ----------
    folder_path : str, optional
        Path to a folder containing files.
    file_list : list, optional
        Explicit list of file paths to process.
    exclude_files : list, optional
        List of file paths to exclude.

    Returns
    -------
    dict
        Dictionary of sorted file lists keyed by type.
    """

    no_file_extension = []
    dicom_files = []
    mhd_files = []
    raw_files = []
    nifti_files = []
    stl_files = []
    vtk_files = []
    mf3_files = []

    if not exclude_files:
        exclude_files = []

    if file_list is None:
        file_list = []
        for root, dirs, files in os.walk(folder_path):
            if files:
                for name in files:
                    file_list += [os.path.join(root, name)]

    for filepath in file_list:
        if filepath not in exclude_files:
            filename, file_extension = os.path.splitext(filepath)

            if file_extension == '.dcm':
                dicom_files.append(filepath)

            elif file_extension == '.mhd':
                mhd_files.append(filepath)

            elif file_extension == '.raw':
                raw_files.append(filepath)

            elif file_extension == '.gz':
                if filepath[-6:] == 'nii.gz':
                    nifti_files.append(filepath)

            elif file_extension == '.stl':
                stl_files.append(filepath)

            elif file_extension == '.vtk':
                vtk_files.append(filepath)

            elif file_extension == '.3mf':
                mf3_files.append(filepath)

            elif file_extension == '':
                no_file_extension.append(filepath)

    files = {'Dicom': dicom_files,
             'MHD': mhd_files,
             'Raw': raw_files,
             'Nifti': nifti_files,
             'Stl': stl_files,
             'Vtk': vtk_files,
             '3mf': mf3_files,
             'NoExtension': no_file_extension}

    return files


def read_dicoms(folder_path=None, file_list=None, exclude_files=None, only_tags=False, only_modality=None,
                only_load_roi_names=None, clear=True):
    """
    Load DICOM files from a folder or file list using DicomReader.

    Parameters
    ----------
    folder_path : str, optional
        Path to folder containing DICOM files.
    file_list : list, optional
        Explicit list of file paths to read.
    exclude_files : list, optional
        List of file paths to exclude.
    only_tags : bool, optional
        If True, only read DICOM metadata tags.
    only_modality : list, optional
        List of DICOM modalities to include (e.g., CT, MR, PT, RTSTRUCT).
    only_load_roi_names : list, optional
        List of ROI names to load.
    clear : bool, optional
        Whether to clear existing DICOM data before loading.
    """
    if only_modality is not None:
        only_modality = only_modality
    else:
        only_modality = ['CT', 'MR', 'PT', 'US', 'DX', 'RF', 'CR', 'RTSTRUCT', 'REG', 'RTDOSE']

    files = None
    if folder_path is not None or file_list is not None:
        files = file_parsar(folder_path=folder_path, file_list=file_list, exclude_files=exclude_files)

    dicom_reader = DicomReader(files, only_tags, only_modality, only_load_roi_names, clear)
    dicom_reader.load()


def read_3mf(file=None, roi_name=None):
    """
    Load 3MF file using ThreeMfReader.

    Parameters
    ----------
    file : str
        Path to the 3MF file.
    roi_name : str, optional
        Name of the ROI to associate with the 3MF file.
    """
    mf3_reader = ThreeMfReader(file, roi_name=roi_name)
    mf3_reader.load()


def read_mhd(file=None, modality=None, reference_name=None, moving_name=None, roi_name=None, dose=None, dvf=None):
    """
    Load MHD (MetaImage) file using MhdReader.

    Parameters
    ----------
    file : str
        Path to the MHD file.
    modality : str, optional
        Imaging modality (e.g., CT, MR).
    reference_name : str, optional
        Name of the reference image for registration.
    moving_name : str, optional
        Name of the moving image for registration.
    roi_name : str, optional
        ROI name to associate with this image.
    dose : optional
        Dose object to associate.
    dvf : optional
        Deformation vector field to associate.
    """
    if file is not None:
        mhd_reader = MhdReader(file=file, modality=modality, reference_name=reference_name, moving_name=moving_name,
                               roi_name=roi_name, dose=dose, dvf=dvf)
        mhd_reader.load()


# def read_stl(self, files=None, create_image=False, match_image=None):
#     stl_reader = StlReader(self)
#     if files is not None:
#         stl_reader.input_files(files)
#     stl_reader.load()
#
#
# def read_vtk(self, files=None, create_image=False, match_image=None):
#         vtk_reader = VtkReader(self)
#         if files is not None:
#             vtk_reader.input_files(files)
#         vtk_reader.load()


if __name__ == '_main__':
    pass



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:

Structure:

"""
import copy

import vtk
import numpy as np
import pyvista as pv
import SimpleITK as sitk

from scipy.spatial.transform import Rotation

from open3d.geometry import PointCloud
from open3d.utility import Vector3dVector
from open3d.pipelines.registration import (registration_icp, ICPConvergenceCriteria,
                                           TransformationEstimationPointToPlane, TransformationEstimationPointToPoint)


class ICP(object):
    """
    Performs Iterative Closest Point (ICP) registration between a source and a target mesh or point cloud. Supports
    both VTK-based and Open3D-based ICP implementations.
    """
    def __init__(self, source, target, matrix=None):
        """
        Initializes the ICP object.

        Parameters
        ----------
        source : pyvista.PolyData or point cloud
            The moving/source mesh or point cloud to align.
        target : pyvista.PolyData or point cloud
            The reference/target mesh or point cloud.
        matrix : np.ndarray, optional
            Initial 4x4 transformation matrix. Defaults to None (identity).
        """
        self.source = source
        self.target = target

        self.matrix = matrix

        self.icp = None

    def compute_com(self):
        """
        Computes an initial translation by matching the center of mass (COM) of the source and target meshes.
        """
        translation = np.asarray(self.mov.center - self.ref.center)

        self.matrix = np.identity(4)
        self.matrix[:3, 3] = translation

    def compute_vtk(self, distance=1e-5, iterations=1000, landmarks=None, com_matching=True, inverse=False):
        """
        Performs ICP using VTK.

        Parameters
        ----------
        distance : float
            Maximum mean distance for convergence.
        iterations : int
            Maximum number of ICP iterations.
        landmarks : int, optional
            Number of landmarks to sample for ICP. Defaults to 1/10 of target points.
        com_matching : bool
            Whether to start by matching centroids of source and target.
        inverse : bool
            If True, stores the inverse of the resulting transformation matrix.
        """
        if landmarks is None:
            landmarks = int(np.round(len(self.target.points) / 10))

        self.icp = vtk.vtkIterativeClosestPointTransform()
        self.icp.SetSource(self.source)
        self.icp.SetTarget(self.target)
        self.icp.GetLandmarkTransform().SetModeToRigidBody()
        self.icp.SetCheckMeanDistance(1)
        self.icp.SetMeanDistanceModeToRMS()
        if landmarks:
            self.icp.SetMaximumNumberOfLandmarks(landmarks)
        self.icp.SetMaximumMeanDistance(distance)
        self.icp.SetMaximumNumberOfIterations(iterations)
        if com_matching:
            self.icp.SetStartByMatchingCentroids(com_matching)
        self.icp.Modified()
        self.icp.Update()

        if inverse:
            self.matrix = np.linalg.inv(pv.array_from_vtkmatrix(self.icp.GetMatrix()))
        else:
            self.matrix = pv.array_from_vtkmatrix(self.icp.GetMatrix())

    def compute_o3d(self, distance=10, iterations=1000, rmse=1e-7, fitness=1e-7, method='point', com_matching=True,
                    inverse=False):
        """
        Performs ICP using Open3D.

        Parameters
        ----------
        distance : float
            Maximum correspondence distance.
        iterations : int
            Maximum number of ICP iterations.
        rmse : float
            Convergence threshold for relative RMSE.
        fitness : float
            Convergence threshold for relative fitness.
        method : str
            'point' for point-to-point ICP, 'plane' for point-to-plane ICP.
        com_matching : bool
            Whether to initialize with translation by centers of mass.
        inverse : bool
            If True, stores the inverse of the resulting transformation matrix.
        """
        ref_pcd = PointCloud()
        ref_pcd.points = Vector3dVector(np.asarray(self.source.points))

        mov_pcd = PointCloud()
        mov_pcd.points = Vector3dVector(np.asarray(self.target.points))
        mov_pcd.normals = Vector3dVector(np.asarray(self.target.point_normals))

        initial_transform = np.identity(4)
        if com_matching:
            translation = mov_pcd.get_center() - ref_pcd.get_center()
            initial_transform[:3, 3] = translation

        if method == 'point':
            self.icp = registration_icp(ref_pcd, mov_pcd, distance, initial_transform,
                                        TransformationEstimationPointToPoint(),
                                        ICPConvergenceCriteria(max_iteration=iterations,
                                                               relative_rmse=rmse,
                                                               relative_fitness=fitness))
        else:
            self.icp = registration_icp(ref_pcd, mov_pcd, distance, initial_transform,
                                        TransformationEstimationPointToPlane())

        if inverse:
            self.matrix = np.linalg.inv(self.icp.transformation)
        else:
            self.matrix = self.icp.transformation

    def get_matrix(self):
        """
        Returns the resulting transformation matrix.

        Returns
        -------
        np.ndarray
            4x4 transformation matrix.
        """

        return self.matrix

    def get_correspondence_set(self):
        """
        Returns the set of corresponding point indices (if available).

        Returns
        -------
        np.ndarray or None
            Correspondence set from ICP.
        """
        if hasattr(self.icp, 'correspondence_set'):
            return np.asarray(self.icp.correspondence_set)

        else:
            return None



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:
"""

import numpy as np
import SimpleITK as sitk


class DeformableITK(object):
    """
    Supports multiple registration methods including BSpline, Elastix, Demons, Fast Demons, and Diffeomorphic Demons,
    with optional mask handling and preprocessing.

    All Demons-based methods support optional smoothing, step size, intensity thresholds, iterations, and mask cropping.
    """
    def __init__(self, reference_image=None, moving_image=None, reference_mask=None, moving_mask=None):
        self.reference_image = reference_image
        self.reference_mask = reference_mask
        self.moving_image = moving_image
        self.moving_mask = moving_mask

    def create_sitk_image(self, array, origin, spacing, direction, reference=True, mask=False):
        """
        Converts a NumPy array to a SimpleITK image with proper origin, spacing, and direction.
        """
        image = sitk.GetImageFromArray(array)
        image.SetOrigin(origin)
        image.SetSpacing(spacing)
        image.SetDirection(direction.flatten().astype(np.float64))

        if reference:
            if mask:
                self.reference_mask = image
            else:
                self.reference_image = image
        else:
            if mask:
                self.moving_mask = image
            else:
                self.moving_image = image

    def cross_modality_correction(self):
        """
        Applies gradient magnitude filtering to reduce intensity differences across modalities.
        """
        if self.reference_image is not None:
            self.reference_image = sitk.GradientMagnitude(self.reference_image)

        if self.moving_image is not None:
            self.moving_image = sitk.GradientMagnitude(self.moving_image)

    def blur_mask(self, sigma=2):
        """
        Smooths masks with a Gaussian filter and normalizes values.
        """
        if self.reference_mask is not None:
            mask = sitk.Cast(self.reference_mask, sitk.sitkFloat32)
            blurred_mask = sitk.SmoothingRecursiveGaussian(mask, sigma=sigma)
            min_val = sitk.GetArrayViewFromImage(blurred_mask).min()
            max_val = sitk.GetArrayViewFromImage(blurred_mask).max()
            self.reference_mask = (blurred_mask - min_val) / (max_val - min_val)

        if self.moving_mask is not None:
            mask = sitk.Cast(self.moving_mask, sitk.sitkFloat32)
            blurred_mask = sitk.SmoothingRecursiveGaussian(mask, sigma=sigma)
            min_val = sitk.GetArrayViewFromImage(blurred_mask).min()
            max_val = sitk.GetArrayViewFromImage(blurred_mask).max()
            self.moving_mask = (blurred_mask - min_val) / (max_val - min_val)

    def resample(self):
        """
        Resamples moving image/mask to match the reference image geometry.
        """
        if self.reference_image is not None and self.moving_image is not None:
            self.moving_image = sitk.Resample(self.moving_image,
                                              self.reference_image,
                                              sitk.Transform(),
                                              sitk.sitkLinear,
                                              0.0,
                                              self.moving_image.GetPixelIDValue())

        if self.reference_mask is not None and self.moving_mask is not None:
            self.moving_mask = sitk.Resample(self.moving_mask,
                                             self.reference_mask,
                                             sitk.Transform(),
                                             sitk.sitkLinear,
                                             0.0,
                                             self.moving_mask.GetPixelIDValue())

    def bspline(self, control_spacing=None, mesh_size=None, gradient=1e-5, iterations=100, crop=5):
        """
        Performs B-spline registration with optional mask constraints, returns displacement vector field.
        """
        if crop > 0:
            self.mask_crop(margin=crop)

        fixed = sitk.Cast(self.reference_image, sitk.sitkFloat32)
        moving = sitk.Cast(self.moving_image, sitk.sitkFloat32)

        if control_spacing is None:
            control_spacing = [50.0, 50.0, 50.0]

        if mesh_size is None:
            image_physical_size = [size * spacing for size, spacing in zip(fixed.GetSize(), fixed.GetSpacing())]
            mesh_size = [int(sz / sp) for sz, sp in zip(image_physical_size, control_spacing)]
        bspline_transform = sitk.BSplineTransformInitializer(fixed, mesh_size)

        bspline = sitk.ImageRegistrationMethod()
        bspline.SetMetricAsMeanSquares()
        if self.reference_mask:
            bspline.SetMetricFixedMask(sitk.Cast(self.reference_mask, sitk.sitkFloat32))
        if self.moving_mask:
            bspline.SetMetricMovingMask(sitk.Cast(self.moving_mask, sitk.sitkFloat32))
        bspline.SetOptimizerAsLBFGSB(gradientConvergenceTolerance=gradient, numberOfIterations=iterations)
        bspline.SetInitialTransform(bspline_transform, inPlace=True)
        bspline.SetInterpolator(sitk.sitkLinear)

        final_transform = bspline.Execute(fixed, moving)

        dvf_image = sitk.TransformToDisplacementField(final_transform, sitk.sitkVectorFloat32, fixed.GetSize(),
                                                      fixed.GetOrigin(), fixed.GetSpacing(), fixed.GetDirection())

        return dvf_image

    def elastix(self, parameter=None, metric='Intensity', bins=6, resolution=4, spacing=10, iterations=2000, order=3,
                crop=5):
        """
        Uses Elastix for nonrigid registration with customizable parameters, returns deformation field.
        """
        if crop > 0:
            self.mask_crop(margin=crop)

        fixed = sitk.Cast(self.reference_image, sitk.sitkFloat32)
        moving = sitk.Cast(self.moving_image, sitk.sitkFloat32)

        elastix = sitk.ElastixImageFilter()
        elastix.SetFixedImage(fixed)
        elastix.SetMovingImage(moving)
        if self.reference_mask is not None:
            elastix.SetFixedMask(sitk.Cast(self.reference_mask, sitk.sitkFloat32))
        if self.moving_mask is not None:
            elastix.SetMovingMask(sitk.Cast(self.moving_mask, sitk.sitkFloat32))

        if parameter is None:
            parameter = sitk.GetDefaultParameterMap("nonrigid")
            if metric == 'Intensity':
                parameter["Metric"] = ['AdvancedMeanSquares']
            else:
                parameter["Metric"] = ["AdvancedMattesMutualInformation"]
            parameter["NumberOfHistogramBins"] = [str(bins)]
            parameter["NumberOfResolutions"] = [str(resolution)]
            parameter["FinalGridSpacingInPhysicalUnits"] = [str(spacing)]
            parameter["MaximumNumberOfIterations"] = [str(iterations)]
            parameter["BSplineTransformSplineOrder"] = [str(order)]
            parameter["AutomaticParameterEstimation"] = ["true"]
            parameter["ResultImagePixelType"] = ["float"]
            parameter["WriteResultImage"] = ["true"]

        elastix.SetParameterMap(parameter)
        elastix.Execute()

        transform_param_map = elastix.GetTransformParameterMap()
        transformix = sitk.TransformixImageFilter()
        transformix.SetMovingImage(moving)
        transformix.SetTransformParameterMap(transform_param_map)
        transform_param_map[0]["ComputeDeformationField"] = ["true"]
        transformix.SetTransformParameterMap(transform_param_map)
        transformix.Execute()

        return transformix.GetDeformationField()

    def demons(self, smooth=True, std=1, iterations=50, intensity_threshold=0.001, crop=5):
        """
        Applies the standard Demons algorithm.
        """
        if crop > 0:
            self.mask_crop(margin=crop)

        fixed = sitk.Cast(self.reference_image, sitk.sitkFloat32)
        moving = sitk.Cast(self.moving_image, sitk.sitkFloat32)

        if self.reference_mask is not None:
            fixed = fixed * sitk.Cast(self.reference_mask, sitk.sitkFloat32)
        if self.moving_mask is not None:
            moving = moving * sitk.Cast(self.moving_mask, sitk.sitkFloat32)

        demons = sitk.DemonsRegistrationFilter()
        demons.SetNumberOfIterations(iterations)
        demons.SetStandardDeviations(std)
        demons.SetIntensityDifferenceThreshold(intensity_threshold)
        if smooth:
            demons.SmoothDisplacementFieldOn()
        else:
            demons.SmoothDisplacementFieldOff()

        return demons.Execute(fixed, moving)

    def fast_demons(self, smooth=True, std=1, iterations=50, intensity_threshold=0.001, step=2.0, crop=5):
        """
        Applies Fast Symmetric Forces Demons algorithm.
        """
        if crop > 0:
            self.mask_crop(margin=crop)

        fixed = sitk.Cast(self.reference_image, sitk.sitkFloat32)
        moving = sitk.Cast(self.moving_image, sitk.sitkFloat32)

        if self.reference_mask is not None:
            fixed = fixed * sitk.Cast(self.reference_mask, sitk.sitkFloat32)
        if self.moving_mask is not None:
            moving = moving * sitk.Cast(self.moving_mask, sitk.sitkFloat32)

        demons = sitk.FastSymmetricForcesDemonsRegistrationFilter()
        demons.SetNumberOfIterations(iterations)
        demons.SetStandardDeviations(std)
        demons.SetIntensityDifferenceThreshold(intensity_threshold)
        demons.SetMaximumUpdateStepLength(step)
        if smooth:
            demons.SmoothDisplacementFieldOn()
        else:
            demons.SmoothDisplacementFieldOff()

        return demons.Execute(fixed, moving)

    def diffeomorphic(self, smooth=True, std=1, iterations=50, intensity_threshold=0.001, step=2.0, crop=5):
        """
        Applies Diffeomorphic Demons algorithm.
        """
        if crop > 0:
            self.mask_crop(margin=crop)

        fixed = sitk.Cast(self.reference_image, sitk.sitkFloat32)
        moving = sitk.Cast(self.moving_image, sitk.sitkFloat32)

        if self.reference_mask is not None:
            fixed = fixed * sitk.Cast(self.reference_mask, sitk.sitkFloat32)
        if self.moving_mask is not None:
            moving = moving * sitk.Cast(self.moving_mask, sitk.sitkFloat32)

        demons = sitk.DiffeomorphicDemonsRegistrationFilter()
        demons.SetNumberOfIterations(iterations)
        demons.SetStandardDeviations(std)
        demons.SetIntensityDifferenceThreshold(intensity_threshold)
        demons.SetMaximumUpdateStepLength(step)
        if smooth:
            demons.SmoothDisplacementFieldOn()
        else:
            demons.SmoothDisplacementFieldOff()

        return demons.Execute(fixed, moving)

    def mask_crop(self, margin=5):
        """
        Crops images and masks to the region covering both masks, with optional padding.
        """
        if self.reference_mask is not None and self.moving_mask is not None:
            combined_mask = sitk.Cast((self.reference_mask > 0) | (self.moving_mask > 0), sitk.sitkUInt8)
            label_stats = sitk.LabelShapeStatisticsImageFilter()
            label_stats.Execute(combined_mask)
            bbox = label_stats.GetBoundingBox(1)

            start = np.array(bbox[:3])
            size = np.array(bbox[3:])
            end = start + size

            start_padded = np.maximum(start - margin, 0)
            end_padded = np.minimum(np.array(self.reference_image.GetSize()), end + margin)
            size_padded = end_padded - start_padded

            self.reference_image = sitk.RegionOfInterest(self.reference_image, size_padded.tolist(), start_padded.tolist())
            self.moving_image = sitk.RegionOfInterest(self.moving_image, size_padded.tolist(), start_padded.tolist())

            self.reference_mask = sitk.RegionOfInterest(self.reference_mask, size_padded.tolist(), start_padded.tolist())
            self.moving_mask = sitk.RegionOfInterest(self.moving_mask, size_padded.tolist(), start_padded.tolist())



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:

Structure:

"""

import os
import copy
import datetime

import numpy as np
import pydicom as dicom
from pydicom.uid import generate_uid, UID, ExplicitVRLittleEndian

from ..data import Data
from ..structure import Image


class CreateDicomImage(object):
    def __init__(self, output_dir, data, study=None, series=None, frame=None, origin=None, spacing=None,
                 thickness=None):
        self.output_dir = output_dir
        self.data = data
        self.study = study
        self.series = series
        self.frame = frame
        self.origin = origin
        self.spacing = spacing
        self.thickness = thickness

        self.orientation = [1, 0, 0, 0, 1, 0]

    def set_study(self, study):
        self.study = study

    def set_series(self, series):
        self.series = series

    def set_frame(self, frame):
        self.frame = frame

    def set_origin(self, origin):
        self.origin = origin

    def set_spacing(self, spacing):
        self.spacing = spacing

    def set_thickness(self, thickness):
        self.thickness = thickness

    def run(self):
        if self.study is None:
            self.study = generate_uid()
        if self.series is None:
            self.series = generate_uid()
        if self.frame is None:
            self.frame = generate_uid()
        if self.origin is None:
            self.origin = [0, 0, 0]
        if self.spacing is None:
            self.spacing = [1, 1]
        if self.thickness is None:
            self.thickness = 1

        for ii in range(self.data.shape[0]):
            array = self.data[ii, :, :]

            ds = dicom.Dataset()
            ds.file_meta = dicom.Dataset()
            ds.file_meta.ImplementationClassUID = "1.2.3.4"
            ds.file_meta.MediaStorageSOPClassUID = UID("1.2.840.10008.5.1.4.1.1.2")
            ds.file_meta.MediaStorageSOPInstanceUID = str(10000 + ii)
            ds.file_meta.TransferSyntaxUID = ExplicitVRLittleEndian

            ds.is_little_endian = True
            ds.is_implicit_VR = False

            ds.PatientName = 'ForAI'
            ds.PatientSex = 'M'
            ds.SeriesDescription = 'Fake for AI'
            ds.PatientID = '12345'
            ds.Modality = 'CT'
            ds.StudyDate = str(datetime.date.today()).replace('-', '')
            ds.ContentDate = str(datetime.date.today()).replace('-', '')
            ds.StudyTime = str(10)
            ds.ContentTime = str(10)
            ds.StudyInstanceUID = self.study
            ds.SeriesInstanceUID = self.series
            ds.SOPInstanceUID = UID(str(10000 + ii))
            ds.SOPClassUID = UID("1.2.840.10008.5.1.4.1.1.2")
            ds.StudyID = '100'

            ds.FrameOfReferenceUID = self.frame
            ds.AcquisitionNumber = '1'
            ds.SeriesNumber = '2'
            ds.InstanceNumber = str(ii + 1)
            ds.ImageOrientationPatient = self.orientation
            ds.PixelSpacing = self.spacing
            ds.SliceThickness = self.thickness
            ds.ImagePositionPatient = [self.origin[0], self.origin[1], (self.origin[2] + (ii * self.thickness))]

            ds.SamplesPerPixel = 1
            ds.PhotometricInterpretation = "MONOCHROME2"
            ds.PixelRepresentation = 1
            ds.HighBit = 15
            ds.BitsStored = 16
            ds.BitsAllocated = 16
            ds.Columns = array.shape[0]
            ds.Rows = array.shape[1]
            ds.RescaleIntercept = 0
            ds.RescaleSlope = 1
            ds.PixelData = array.tobytes()

            export_file = os.path.join(self.output_dir, str(ii) + '.dcm')
            ds.save_as(export_file, write_like_original=False)


class CreateImageFromMask(object):
    def __init__(self, array, origin, spacing, image_name, dimensions=None, orientation=None, plane='Axial',
                 description='Mask to Image', modality='CT'):
        self.rois = {}
        self.pois = {}

        self.array = array
        self.spacing = spacing
        self.origin = origin

        self.image_name = image_name

        now = datetime.datetime.now()
        self.date = str(now.year) + str(now.month) + str(now.day)
        if len(str(now.second)) == 1:
            self.time = str(now.hour) + '0' + str(now.second) + '00'
        else:
            self.time = str(now.hour) + str(now.second) + '00'
        self.birthdate = self.date

        self.filepaths = None

        self.plane = plane
        if dimensions is None:
            self.dimensions = array.shape
        else:
            self.dimensions = dimensions

        if orientation is None:
            self.orientation = [1, 0, 0, 0, 1, 0]
        else:
            self.orientation = orientation

        row_direction = self.orientation[:3]
        column_direction = self.orientation[3:6]
        slice_direction = np.cross(row_direction, column_direction)
        self.image_matrix = np.identity(3, dtype=np.float32)
        self.image_matrix[0, :3] = row_direction
        self.image_matrix[1, :3] = column_direction
        self.image_matrix[2, :3] = slice_direction

        self.camera_position = None
        self.unverified = None
        self.skipped_slice = None
        self.sections = None
        self.rgb = False

        self.sops = [generate_uid() for ii in range(self.dimensions[0])]
        self.slice_location = [int(self.dimensions[0] / 2), int(self.dimensions[1] / 2), int(self.dimensions[2] / 2)]

        self.study_uid = generate_uid()
        self.series_uid = generate_uid()
        self.frame_ref = generate_uid()
        self.acq_number = '1'
        self.window = [0, 1]
        self.modality = modality
        sop_class = generate_uid()

        dicoms = []
        for ii in range(self.dimensions[0]):
            ds = dicom.Dataset()
            ds.file_meta = dicom.Dataset()
            ds.file_meta.ImplementationClassUID = "1.2.3.4"
            ds.file_meta.MediaStorageSOPClassUID = UID(sop_class)
            ds.file_meta.MediaStorageSOPInstanceUID = UID(str(self.sops[ii]))
            ds.file_meta.TransferSyntaxUID = ExplicitVRLittleEndian

            ds.is_little_endian = True
            ds.is_implicit_VR = False

            ds.PatientName = 'User^Created^ ^'
            ds.PatientSex = 'M'
            ds.SeriesDescription = description
            ds.PatientID = 'User^Created^ ^'
            ds.Modality = modality
            ds.StudyDate = self.date
            ds.ContentDate = self.date
            ds.StudyTime = self.time
            ds.ContentTime = self.time
            ds.StudyInstanceUID = self.study_uid
            ds.SeriesInstanceUID = self.series_uid
            ds.SOPInstanceUID = UID(str(self.sops[ii]))
            ds.SOPClassUID = UID(str(sop_class))
            ds.StudyID = '1'

            ds.FrameOfReferenceUID = self.frame_ref
            ds.AcquisitionNumber = self.acq_number
            ds.SeriesNumber = '1'
            ds.InstanceNumber = str(ii)
            ds.ImageOrientationPatient = list(self.orientation[:6])
            ds.PixelSpacing = list(spacing[:2])
            ds.SliceThickness = spacing[2]

            position = self.compute_position(ii)
            ds.ImagePositionPatient = [position[0], position[1], position[2]]

            ds.SamplesPerPixel = 1
            ds.PhotometricInterpretation = "MONOCHROME2"
            ds.PixelRepresentation = 1
            ds.HighBit = 15
            ds.BitsStored = 16
            ds.BitsAllocated = 16
            ds.Columns = array.shape[1]
            ds.Rows = array.shape[2]
            ds.RescaleIntercept = 0
            ds.RescaleSlope = 1

            dicoms += [ds]

        self.image_set = dicoms

    def add_image(self):
        Data.image[self.image_name] = Image(self)
        Data.image_list += [self.image_name]

    def add_mesh_roi(self, mesh, roi_name):
        Data.image[self.image_name].create_roi(self, name=roi_name, color=[0, 0, 255], visible=False, filepath=None)
        self.rois[roi_name].mesh = mesh
        
        self.rois[roi_name].volume = mesh.volume
        self.rois[roi_name].com = mesh.center
        self.rois[roi_name].bounds = mesh.bounds

    def compute_position(self, z):
        matrix = copy.deepcopy(self.image_matrix)

        pixel_to_position_matrix = np.identity(4, dtype=np.float32)
        pixel_to_position_matrix[:3, 0] = matrix[0, :] * self.spacing[0]
        pixel_to_position_matrix[:3, 1] = matrix[1, :] * self.spacing[1]
        pixel_to_position_matrix[:3, 2] = matrix[2, :] * self.spacing[2]
        pixel_to_position_matrix[:3, 3] = self.origin

        location = np.asarray([0, 0, z, 1])

        return location.dot(pixel_to_position_matrix.T)[:3]


"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:
"""

import numpy as np
import SimpleITK as sitk


def euler_transform(matrix=None, angles=None, translation=None, rotation_center=None, zyx=False):
    """
    Creates a 3D Euler transformation that can combine rotation, translation, and a rotation center. Useful for rigid or
    semi-rigid registration in 3D.
    """
    transform = sitk.Euler3DTransform()

    if angles is not None:
        rotation = [angles[0] * np.pi / 180, angles[1] * np.pi / 180, angles[2] * np.pi / 180]
        transform.SetRotation(rotation[0], rotation[1], rotation[2])

    if matrix is not None:
        transform.SetMatrix(matrix[:3, :3].flatten().astype(np.double))

    if translation is not None:
        transform.SetTranslation(translation)

    if rotation_center is not None:
        transform.SetCenter(rotation_center)

    if zyx:
        transform.SetComputeZYX(True)

    return transform



"""
Morfeus lab
The University of Texas
MD Anderson Cancer Center
Author - Caleb O'Connor
Email - csoconnor@mdanderson.org

Description:

Structure:

"""

import os

import numpy as np

import SimpleITK as sitk

from ..data import Data
from ..structure.deformable import Deformable
from ..utils.creation import CreateImageFromMask


class MhdReader(object):
    """
    Class for reading .mhd (MetaImage) files and optionally creating images, ROIs, doses, or deformable vector fields.
    """
    def __init__(self, file, modality=None, reference_name=None, moving_name=None, roi_name=None, dose=None, dvf=None):
        """
        Initialize the reader.

        Parameters:
        - file: path to the .mhd file
        - modality: type of image (e.g., 'CT', 'MR')
        - reference_name: reference image name for DVF registration
        - moving_name: moving image name for DVF registration
        - roi_name: name of ROI to create
        - dose: dose object (optional)
        - dvf: deformable vector field (optional)
        """
        self.file = file
        self.modality = modality
        self.reference_name = reference_name
        self.moving_name = moving_name
        self.roi_name = roi_name
        self.dose = dose
        self.dvf = dvf

        self.mhd = None

    def load(self):
        """
        Load the .mhd file and decide what to create based on provided attributes.
        """
        self.mhd = sitk.ReadImage(self.file)

        if self.reference_name is not None:
            if self.dvf is not None and self.moving_name is not None:
                self.create_dvf()

            elif self.dose is not None:
                pass

            elif self.roi_name is not None:
                pass

        else:
            self.create_image()

    def create_image(self):
        """
        Converts the loaded .mhd file into an internal Image object and registers it in the Data structure.
        """
        if self.modality is None:
            filename = os.path.basename(self.file)
            image_name = os.path.splitext(filename)[0]
            self.modality = 'CT'
        else:
            idx = len(Data.image_list)
            if idx < 9:
                image_name = self.modality + ' 0' + str(1 + idx)
            else:
                image_name = self.modality + ' ' + str(1 + idx)

        dimensions = np.flip(np.asarray(self.mhd.GetSize()))
        orientation = np.asarray(self.mhd.GetDirection())
        origin = np.asarray(self.mhd.GetOrigin())
        spacing = np.asarray(self.mhd.GetSpacing())

        array = sitk.GetArrayFromImage(self.mhd)

        create = CreateImageFromMask(array, origin, spacing, image_name, dimensions=dimensions, orientation=orientation,
                                     plane='Axial', description='Mhd to Image', modality=self.modality)
        create.add_image()

    def create_roi(self):
        pass

    def create_dose(self):
        pass

    def create_dvf(self):
        """
        Create a deformable vector field (DVF) from the loaded .mhd file. Handles naming conflicts and registers the
        DVF in the Data structure.
        """
        registration_name = 'DVF_' + self.reference_name + '_' + self.moving_name
        if registration_name in Data.deformable_list:
            n = 0
            while n > -1:
                n += 1
                new_name = copy.deepcopy(registration_name + '_' + str(n))
                if new_name not in Data.deformable_list:
                    registration_name = new_name
                    n = -100

        dimensions = np.flip(np.asarray(self.mhd.GetSize()))
        dvf_matrix = np.asarray(self.mhd.GetDirection()).reshape(3, 3)
        origin = np.asarray(self.mhd.GetOrigin())
        spacing = np.asarray(self.mhd.GetSpacing())

        array = sitk.GetArrayFromImage(self.mhd)

        Deformable(array, origin, spacing, dimensions, dvf_matrix=dvf_matrix, registration_name=registration_name,
                   reference_name=self.reference_name, moving_name=self.moving_name)
